# -*- coding: utf-8 -*-

"""Supports the Ion Velocity Meter (IVM)
onboard the Ionospheric Connections (ICON) Explorer.

Properties
----------
platform
    'icon'
name
    'ivm'
tag
    None supported
sat_id
    'a' or 'b'


Warnings
--------
- No download routine as ICON has not yet been launched
- Data not yet publicly available


Example
-------
::

    import pysat
    ivm = pysat.Instrument(platform='icon', name='ivm', sat_id='a')
    ivm.download(dt.datetime(2020, 1, 1), dt.datetime(2020, 1, 31))
    ivm.load(2020, 1)

By default, pysat removes the ICON level tags from variable names, ie,
ICON_L27_Ion_Density becomes Ion_Density.  To retain the original names, use
::

    ivm = pysat.Instrument(platform='icon', name='ivm', sat_id='a',
                           keep_original_names=True)


Author
------
R. A. Stoneback

"""

from __future__ import print_function
from __future__ import absolute_import

import datetime as dt
import functools
import logging
import numpy as np

import pysat
from pysat.instruments.methods import general as mm_gen
from pysat.instruments.methods import icon as mm_icon


logger = logging.getLogger(__name__)

platform = 'icon'
name = 'ivm'
tags = {'': 'Level 2 public geophysical data'}
# Note for developers: IVM-A and IVM-B face in opposite directions, and only
# one is expected to have geophysical data at a given time depending on ram
# direction. In general, IVM-A is operational when the remote instruments face
# northward, and IVM-B when the remote instruments face southward. IVM-B data
# is not available as of Jun 26 2020, as this mode has not yet been engaged.
# Bypassing tests and warning checks via the password_req flag
sat_ids = {'a': [''],
           'b': ['']}
_test_dates = {'a': {'': dt.datetime(2020, 1, 1)},
               'b': {'': dt.datetime(2020, 1, 1)}}  # IVM-B not yet engaged
_test_download_travis = {'a': {kk: False for kk in tags.keys()}}
_test_download = {'b': {kk: False for kk in tags.keys()}}
_password_req = {'b': {kk: True for kk in tags.keys()}}

aname = ''.join(('ICON_L2-7_IVM-A_{year:04d}-{month:02d}-{day:02d}_',
                 'v{version:02d}r{revision:03d}.NC'))
bname = ''.join(('ICON_L2-7_IVM-B_{year:04d}-{month:02d}-{day:02d}_',
                 'v{version:02d}r{revision:03d}.NC'))
supported_tags = {'a': {'': aname},
                  'b': {'': bname}}

# use the general methods list files routine
list_files = functools.partial(mm_gen.list_files,
                               supported_tags=supported_tags)

# support download routine
basic_tag_a = {'dir': '/pub/LEVEL.2/IVM-A',
               'remote_fname': 'Data/' + aname}
basic_tag_b = {'dir': '/pub/LEVEL.2/IVM-B',
               'remote_fname': 'Data/' + bname}

download_tags = {'a': {'': basic_tag_a},
                 'b': {'': basic_tag_b}}
download = functools.partial(mm_icon.ssl_download, supported_tags=download_tags)

# support listing files on SSL
list_remote_files = functools.partial(mm_icon.list_remote_files,
                                      supported_tags=download_tags)


def init(self):
    """Initializes the Instrument object with instrument specific values.

    Runs once upon instantiation.

    Parameters
    -----------
    inst : (pysat.Instrument)
        Instrument class object

    """

    logger.info(mm_icon.ackn_str)
    self.meta.acknowledgements = mm_icon.ackn_str
    self.meta.references = ''.join((mm_icon.refs['mission'],
                                    mm_icon.refs['ivm']))

    pass


def default(inst):
    """Default routine to be applied when loading data. Removes variable
    preambles.

    Parameters
    -----------
    inst : (pysat.Instrument)
        Instrument class object

    """

    if not inst.kwargs['keep_original_names']:
        mm_gen.remove_leading_text(inst, target='ICON_L27_')


def load(fnames, tag=None, sat_id=None, keep_original_names=False):
    """Loads ICON IVM data using pysat into pandas.

    This routine is called as needed by pysat. It is not intended
    for direct user interaction.

    Parameters
    ----------
    fnames : array-like
        iterable of filename strings, full path, to data files to be loaded.
        This input is nominally provided by pysat itself.
    tag : string
        tag name used to identify particular data set to be loaded.
        This input is nominally provided by pysat itself.
    sat_id : string
        Satellite ID used to identify particular data set to be loaded.
        This input is nominally provided by pysat itself.
    keep_original_names : boolean
        if True then the names as given in the netCDF ICON file
        will be used as is. If False, a preamble is removed.

    Returns
    -------
    data, metadata
        Data and Metadata are formatted for pysat. Data is a pandas
        DataFrame while metadata is a pysat.Meta instance.

    Note
    ----
    Any additional keyword arguments passed to pysat.Instrument
    upon instantiation are passed along to this routine.

    Examples
    --------
    ::
        inst = pysat.Instrument('icon', 'ivm', sat_id='a', tag='')
        inst.load(2020, 1)

    """

    return pysat.utils.load_netcdf4(fnames, epoch_name='Epoch',
                                    units_label='Units',
                                    name_label='Long_Name',
                                    notes_label='Var_Notes',
                                    desc_label='CatDesc',
                                    plot_label='FieldNam',
                                    axis_label='LablAxis',
                                    scale_label='ScaleTyp',
                                    min_label='ValidMin',
                                    max_label='ValidMax',
                                    fill_label='FillVal')


def clean(inst):
    """Provides data cleaning based upon clean_level.

    clean_level is set upon Instrument instantiation to
    one of the following:

    'Clean'
    'Dusty'
    'Dirty'
    'None'

    Routine is called by pysat, and not by the end user directly.

    Parameters
    -----------
    inst : (pysat.Instrument)
        Instrument class object, whose attribute clean_level is used to return
        the desired level of data selectivity.

    Note
    ----
        Supports 'clean', 'dusty', 'dirty', 'none'

    """

    # IVM variable groupings
    drift_variables = ['Ion_Velocity_X', 'Ion_Velocity_Zonal',
                       'Ion_Velocity_Meridional',
                       'Ion_Velocity_Field_Aligned']
    cross_drift_variables = ['Ion_Velocity_Z', 'Ion_Velocity_Y']
    rpa_variables = ['Ion_Temperature', 'Ion_Density',
                     'Fractional_Ion_Density_H',
                     'Fractional_Ion_Density_O']
    if 'RPA_Flag' in inst.variables:
        rpa_flag = 'RPA_Flag'
        dm_flag = 'DM_Flag'
    else:
        rpa_flag = 'ICON_L27_RPA_Flag'
        dm_flag = 'ICON_L27_DM_Flag'
        drift_variables = ['ICON_L27_' + x for x in drift_variables]
        cross_drift_variables = ['ICON_L27_' + x for x in cross_drift_variables]
        rpa_variables = ['ICON_L27_' + x for x in rpa_variables]

    if inst.clean_level in ['clean', 'dusty']:
        # remove drift values impacted by RPA
        idx, = np.where(inst[rpa_flag] >= 1)
        for var in drift_variables:
            inst[idx, var] = np.nan
        # DM values
        idx, = np.where(inst[dm_flag] >= 1)
        for var in cross_drift_variables:
            inst[idx, var] = np.nan

    if inst.clean_level == 'clean':
        # other RPA parameters
        idx, = np.where(inst[rpa_flag] >= 2)
        for var in rpa_variables:
            inst[idx, var] = np.nan

    return



from nose.tools import raises
import warnings

import pysat
from pysat import model_utils as mu


class TestBasics():
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument(platform='pysat',
                                         name='testing',
                                         clean_level='clean')
        self.start = pysat.datetime(2009, 1, 1)
        self.stop = pysat.datetime(2009, 1, 1)

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.start, self.stop

    @raises(ValueError)
    def test_collect_inst_model_pairs_wo_date(self):
        """Try to run without start or stop dates"""
        mu.collect_inst_model_pairs(inst=self.testInst)

    @raises(ValueError)
    def test_collect_inst_model_pairs_wo_inst(self):
        """Try to run without an instrument"""
        mu.collect_inst_model_pairs(start=self.start, stop=self.stop)

    @raises(ValueError)
    def test_collect_inst_model_pairs_wo_model(self):
        """Try to run without a model"""
        mu.collect_inst_model_pairs(start=self.start, stop=self.stop,
                                    inst=self.testInst)


class TestDeprecation():

    def setup(self):
        """Runs before every method to create a clean testing setup"""
        warnings.simplefilter("always")

    def teardown(self):
        """Runs after every method to clean up previous testing"""

    def test_satellite_view_through_model_deprecation(self):
        """Test if satellite_view_through_model is deprecated"""

        with warnings.catch_warnings(record=True) as war:
            try:
                mu.satellite_view_through_model(None, None, None, None)
            except TypeError:
                # Setting inst to None should produce a TypeError after
                # warning is generated
                pass

        assert len(war) >= 1
        assert war[0].category == DeprecationWarning

    def test_collect_inst_model_pairs_deprecation(self):
        """Test if collect_inst_model_pairs is deprecated"""

        with warnings.catch_warnings(record=True) as war:
            try:
                mu.collect_inst_model_pairs(inst=None)
            except ValueError:
                # Setting inst to None should produce a ValueError after
                # warning is generated
                pass

        assert len(war) >= 1
        assert war[0].category == DeprecationWarning

    def test_compare_model_and_inst_deprecation(self):
        """Test if compare_model_and_inst is deprecated"""

        with warnings.catch_warnings(record=True) as war:
            try:
                mu.compare_model_and_inst(pairs=None)
            except ValueError:
                # Setting pairs to None should produce a ValueError after
                # warning is generated
                pass

        assert len(war) >= 1
        assert war[0].category == DeprecationWarning

    def test_extract_modelled_observations_deprecation(self):
        """Test if extract_modelled_observations is deprecated"""

        with warnings.catch_warnings(record=True) as war:
            try:
                mu.extract_modelled_observations(inst=None)
            except ValueError:
                # Setting inst to None should produce a ValueError after
                # warning is generated
                pass

        assert len(war) >= 1
        assert war[0].category == DeprecationWarning



# -*- coding: utf-8 -*-
"""Supports the Far Ultraviolet (FUV) imager onboard the Ionospheric
CONnection Explorer (ICON) satellite.  Accesses local data in
netCDF format.

Properties
----------
platform
    'icon'
name
    'fuv'
tag
    None supported


Warnings
--------
- The cleaning parameters for the instrument are still under development.
- Only supports level-2 data.


Example
-------
::

    import pysat
    fuv = pysat.Instrument(platform='icon', name='fuv', tag='day')
    fuv.download(dt.datetime(2020, 1, 1), dt.datetime(2020, 1, 31))
    fuv.load(2020, 1)

By default, pysat removes the ICON level tags from variable names, ie,
ICON_L27_Ion_Density becomes Ion_Density.  To retain the original names, use
::

    fuv = pysat.Instrument(platform='icon', name='fuv', tag=day',
                           keep_original_names=True)


Authors
---------
Originated from EUV support.
Jeff Klenzing, Mar 17, 2018, Goddard Space Flight Center
Russell Stoneback, Mar 23, 2018, University of Texas at Dallas
Conversion to FUV, Oct 8th, 2028, University of Texas at Dallas

"""

from __future__ import print_function
from __future__ import absolute_import

import datetime as dt
import functools
import logging
import warnings

import pysat
from pysat.instruments.methods import general as mm_gen
from pysat.instruments.methods import icon as mm_icon


logger = logging.getLogger(__name__)

platform = 'icon'
name = 'fuv'
tags = {'day': 'Level 2 daytime O/N2',
        'night': 'Level 2 nighttime O profile'}
sat_ids = {'': ['day', 'night']}
_test_dates = {'': {kk: dt.datetime(2020, 1, 1) for kk in tags.keys()}}
_test_download_travis = {'': {kk: False for kk in tags.keys()}}
pandas_format = False

fname24 = ''.join(('ICON_L2-4_FUV_Day_{year:04d}-{month:02d}-{day:02d}_',
                   'v{version:02d}r{revision:03d}.NC'))
fname25 = ''.join(('ICON_L2-5_FUV_Night_{year:04d}-{month:02d}-{day:02d}_',
                   'v{version:02d}r{revision:03d}.NC'))
supported_tags = {'': {'day': fname24,
                       'night': fname25}}

# use the standard methods list files routine
list_files = functools.partial(mm_gen.list_files,
                               supported_tags=supported_tags)

# support download routine
basic_tag24 = {'dir': '/pub/LEVEL.2/FUV',
               'remote_fname': fname24}
basic_tag25 = {'dir': '/pub/LEVEL.2/FUV',
               'remote_fname': fname25}

download_tags = {'': {'day': basic_tag24,
                      'night': basic_tag25}}

download = functools.partial(mm_icon.ssl_download, supported_tags=download_tags)

# support listing files on SSL
list_remote_files = functools.partial(mm_icon.list_remote_files,
                                      supported_tags=download_tags)


def init(self):
    """Initializes the Instrument object with instrument specific values.

    Runs once upon instantiation.

    Parameters
    -----------
    inst : (pysat.Instrument)
        Instrument class object

    """

    logger.info(mm_icon.ackn_str)
    self.meta.acknowledgements = mm_icon.ackn_str
    self.meta.references = ''.join((mm_icon.refs['mission'],
                                    mm_icon.refs['fuv']))

    pass


def default(inst):
    """Default routine to be applied when loading data. Adjusts epoch timestamps
    to datetimes and removes variable preambles.

    Parameters
    -----------
    inst : (pysat.Instrument)
        Instrument class object


    """

    mm_gen.convert_timestamp_to_datetime(inst, sec_mult=1.0e-3)
    if not inst.kwargs['keep_original_names']:
        remove_preamble(inst)


def remove_preamble(inst):
    """Removes preambles in variable names"""

    target = {'day': 'ICON_L24_',
              'night': 'ICON_L25_'}
    mm_gen.remove_leading_text(inst, target=target[inst.tag])


def load(fnames, tag=None, sat_id=None, keep_original_names=False):
    """Loads ICON FUV data using pysat into pandas.

    This routine is called as needed by pysat. It is not intended
    for direct user interaction.

    Parameters
    ----------
    fnames : array-like
        iterable of filename strings, full path, to data files to be loaded.
        This input is nominally provided by pysat itself.
    tag : string
        tag name used to identify particular data set to be loaded.
        This input is nominally provided by pysat itself.
    sat_id : string
        Satellite ID used to identify particular data set to be loaded.
        This input is nominally provided by pysat itself.
    keep_original_names : boolean
        if True then the names as given in the netCDF ICON file
        will be used as is. If False, a preamble is removed.

    Returns
    -------
    data, metadata
        Data and Metadata are formatted for pysat. Data is a pandas
        DataFrame while metadata is a pysat.Meta instance.

    Note
    ----
    Any additional keyword arguments passed to pysat.Instrument
    upon instantiation are passed along to this routine.

    Examples
    --------
    ::
        inst = pysat.Instrument('icon', 'fuv')
        inst.load(2020, 1)

    """

    return pysat.utils.load_netcdf4(fnames, epoch_name='Epoch',
                                    units_label='Units',
                                    name_label='Long_Name',
                                    notes_label='Var_Notes',
                                    desc_label='CatDesc',
                                    plot_label='FieldNam',
                                    axis_label='LablAxis',
                                    scale_label='ScaleTyp',
                                    min_label='ValidMin',
                                    max_label='ValidMax',
                                    fill_label='FillVal',
                                    pandas_format=pandas_format)


def clean(inst):
    """Provides data cleaning based upon clean_level.

    clean_level is set upon Instrument instantiation to
    one of the following:

    'Clean'
    'Dusty'
    'Dirty'
    'None'

    Routine is called by pysat, and not by the end user directly.

    Parameters
    -----------
    inst : (pysat.Instrument)
        Instrument class object, whose attribute clean_level is used to return
        the desired level of data selectivity.

    Note
    ----
        Supports 'clean', 'dusty', 'dirty', 'none'

    """

    warnings.warn("Cleaning actions for ICON FUV are not yet defined.")
    return



# -*- coding: utf-8 -*-
"""Provides default routines for integrating NASA CDAWeb instruments into
pysat. Adding new CDAWeb datasets should only require mininal user
intervention.

"""

from __future__ import absolute_import, division, print_function
import logging
import sys
import warnings

import pandas as pds

import pysat
from pysat.instruments.methods import general as mm_gen

logger = logging.getLogger(__name__)


def list_files(tag=None, sat_id=None, data_path=None, format_str=None,
               supported_tags=None, fake_daily_files_from_monthly=False,
               two_digit_year_break=None):
    """Return a Pandas Series of every file for chosen satellite data.

    .. deprecated:: 2.2.0
      `list_files` will be removed in pysat 3.0.0, it will be replaced by the
      copy in instruments.methods.general

    This routine is intended to be used by pysat instrument modules supporting
    a particular NASA CDAWeb dataset.

    Parameters
    -----------
    tag : (string or NoneType)
        Denotes type of file to load.  Accepted types are <tag strings>.
        (default=None)
    sat_id : (string or NoneType)
        Specifies the satellite ID for a constellation.  Not used.
        (default=None)
    data_path : (string or NoneType)
        Path to data directory.  If None is specified, the value previously
        set in Instrument.files.data_path is used.  (default=None)
    format_str : (string or NoneType)
        User specified file format.  If None is specified, the default
        formats associated with the supplied tags are used. (default=None)
    supported_tags : (dict or NoneType)
        keys are sat_id, each containing a dict keyed by tag
        where the values file format template strings. (default=None)
    fake_daily_files_from_monthly : (bool)
        Some CDAWeb instrument data files are stored by month, interfering
        with pysat's functionality of loading by day. This flag, when true,
        appends daily dates to monthly files internally. These dates are
        used by load routine in this module to provide data by day.
    two_digit_year_break : (int)
        If filenames only store two digits for the year, then
        '1900' will be added for years >= two_digit_year_break
        and '2000' will be added for years < two_digit_year_break.

    Returns
    --------
    pysat.Files.from_os : (pysat._files.Files)
        A class containing the verified available files

    Examples
    --------
    ::

        fname = 'cnofs_vefi_bfield_1sec_{year:04d}{month:02d}{day:02d}_v05.cdf'
        supported_tags = {'dc_b': fname}
        list_files = functools.partial(nasa_cdaweb.list_files,
                                       supported_tags=supported_tags)

        fname = 'cnofs_cindi_ivm_500ms_{year:4d}{month:02d}{day:02d}_v01.cdf'
        supported_tags = {'': fname}
        list_files = functools.partial(cdw.list_files,
                                       supported_tags=supported_tags)

    """

    warnings.warn(' '.join(["methods.nasa_cdaweb.list_files has been",
                            "deprecated and will be removed in pysat 3.0.0.",
                            "Please use methods.general.list_files instead"]),
                  DeprecationWarning, stacklevel=2)

    out = mm_gen.list_files(tag=tag, sat_id=sat_id, data_path=data_path,
                            format_str=format_str,
                            supported_tags=supported_tags,
                            fake_daily_files_from_monthly=fake_daily_files_from_monthly,
                            two_digit_year_break=two_digit_year_break)
    return out


def load(fnames, tag=None, sat_id=None,
         fake_daily_files_from_monthly=False,
         flatten_twod=True):
    """Load NASA CDAWeb CDF files.

    This routine is intended to be used by pysat instrument modules supporting
    a particular NASA CDAWeb dataset.

    Parameters
    ------------
    fnames : (pandas.Series)
        Series of filenames
    tag : (str or NoneType)
        tag or None (default=None)
    sat_id : (str or NoneType)
        satellite id or None (default=None)
    fake_daily_files_from_monthly : bool
        Some CDAWeb instrument data files are stored by month, interfering
        with pysat's functionality of loading by day. This flag, when true,
        parses of daily dates to monthly files that were added internally
        by the list_files routine, when flagged. These dates are
        used here to provide data by day.
    flatted_twod : bool
        Flattens 2D data into different columns of root DataFrame rather
        than produce a Series of DataFrames

    Returns
    ---------
    data : (pandas.DataFrame)
        Object containing satellite data
    meta : (pysat.Meta)
        Object containing metadata such as column names and units

    Examples
    --------
    ::

        # within the new instrument module, at the top level define
        # a new variable named load, and set it equal to this load method
        # code below taken from cnofs_ivm.py.

        # support load routine
        # use the default CDAWeb method
        load = cdw.load


    """

    import pysatCDF

    if len(fnames) <= 0:
        return pysat.DataFrame(None), None
    else:
        # going to use pysatCDF to load the CDF and format
        # data and metadata for pysat using some assumptions.
        # Depending upon your needs the resulting pandas DataFrame may
        # need modification
        # currently only loads one file, which handles more situations via
        # pysat than you may initially think

        if fake_daily_files_from_monthly:
            # parse out date from filename
            fname = fnames[0][0:-11]
            date = pysat.datetime.strptime(fnames[0][-10:], '%Y-%m-%d')
            with pysatCDF.CDF(fname) as cdf:
                # convert data to pysat format
                data, meta = cdf.to_pysat(flatten_twod=flatten_twod)
                # select data from monthly
                data = data.loc[date:date+pds.DateOffset(days=1)
                                - pds.DateOffset(microseconds=1), :]
                return data, meta
        else:
            # basic data return
            with pysatCDF.CDF(fnames[0]) as cdf:
                return cdf.to_pysat(flatten_twod=flatten_twod)


def download(supported_tags, date_array, tag, sat_id,
             remote_site='https://cdaweb.gsfc.nasa.gov',
             data_path=None, user=None, password=None,
             fake_daily_files_from_monthly=False,
             multi_file_day=False):
    """Routine to download NASA CDAWeb CDF data.

    This routine is intended to be used by pysat instrument modules supporting
    a particular NASA CDAWeb dataset.

    Parameters
    -----------
    supported_tags : dict
        dict of dicts. Keys are supported tag names for download. Value is
        a dict with 'dir', 'remote_fname', 'local_fname'. Inteded to be
        pre-set with functools.partial then assigned to new instrument code.
    date_array : array_like
        Array of datetimes to download data for. Provided by pysat.
    tag : str or NoneType (None)
        tag or None
    sat_id : (str or NoneType)
        satellite id or None (default=None)
    remote_site : (string or NoneType)
        Remote site to download data from
        (default='https://cdaweb.gsfc.nasa.gov')
    data_path : (string or NoneType)
        Path to data directory.  If None is specified, the value previously
        set in Instrument.files.data_path is used.  (default=None)
    user : (string or NoneType)
        Username to be passed along to resource with relevant data.
        (default=None)
    password : (string or NoneType)
        User password to be passed along to resource with relevant data.
        (default=None)
    fake_daily_files_from_monthly : bool
        Some CDAWeb instrument data files are stored by month. This flag,
        when true, accomodates this reality with user feedback on a monthly
        time frame.

    Returns
    --------
    Void : (NoneType)
        Downloads data to disk.

    Examples
    --------
    ::

        # download support added to cnofs_vefi.py using code below
        rn = '{year:4d}/cnofs_vefi_bfield_1sec_{year:4d}{month:02d}{day:02d}'+
            '_v05.cdf'
        ln = 'cnofs_vefi_bfield_1sec_{year:4d}{month:02d}{day:02d}_v05.cdf'
        dc_b_tag = {'dir':'/pub/data/cnofs/vefi/bfield_1sec',
                    'remote_fname': rn,
                    'local_fname': ln}
        supported_tags = {'dc_b': dc_b_tag}

        download = functools.partial(nasa_cdaweb.download,
                                     supported_tags=supported_tags)

    """

    import os
    import requests

    try:
        inst_dict = supported_tags[sat_id][tag]
    except KeyError:
        raise ValueError('Tag name unknown.')

    # path to relevant file on CDAWeb
    remote_url = remote_site + inst_dict['dir']

    # naming scheme for files on the CDAWeb server
    remote_fname = inst_dict['remote_fname']

    # naming scheme for local files, should be closely related
    # to CDAWeb scheme, though directory structures may be reduced
    # if desired
    local_fname = inst_dict['local_fname']

    if not multi_file_day:
        # Get list of files from server
        remote_files = list_remote_files(tag=tag, sat_id=sat_id,
                                         remote_site=remote_site,
                                         supported_tags=supported_tags)
        # Find only requested files that exist remotely
        date_array = pds.DatetimeIndex(list(set(remote_files.index)
                                            & set(date_array))).sort_values()

    for date in date_array:
        # format files for specific dates and download location
        formatted_remote_fname = remote_fname.format(year=date.year,
                                                     month=date.month,
                                                     day=date.day,
                                                     hour=date.hour,
                                                     min=date.minute,
                                                     sec=date.second)
        formatted_local_fname = local_fname.format(year=date.year,
                                                   month=date.month,
                                                   day=date.day,
                                                   hour=date.hour,
                                                   min=date.minute,
                                                   sec=date.second)
        saved_local_fname = os.path.join(data_path, formatted_local_fname)

        # perform download
        if not multi_file_day:
            # standard download
            try:
                logger.info(' '.join(('Attempting to download file for',
                                      date.strftime('%d %B %Y'))))
                sys.stdout.flush()
                remote_path = '/'.join((remote_url.strip('/'),
                                        formatted_remote_fname))
                req = requests.get(remote_path)
                if req.status_code != 404:
                    open(saved_local_fname, 'wb').write(req.content)
                    logger.info('Finished.')
                else:
                    logger.info(' '.join(('File not available for',
                                          date.strftime('%d %B %Y'))))
            except requests.exceptions.RequestException as exception:
                logger.info(' '.join((exception, '- File not available for',
                                      date.strftime('%d %B %Y'))))
        else:
            try:
                logger.info(' '.join(('Attempting to download files for',
                                      date.strftime('%d %B %Y'))))
                sys.stdout.flush()
                remote_files = list_remote_files(tag=tag, sat_id=sat_id,
                                                 remote_site=remote_site,
                                                 supported_tags=supported_tags,
                                                 year=date.year,
                                                 month=date.month,
                                                 day=date.day)

                # Get the files
                i = 0
                n = len(remote_files.values)
                for remote_file in remote_files.values:
                    remote_dir = os.path.split(formatted_remote_fname)[0]
                    remote_file_path = '/'.join((remote_url.strip('/'),
                                                 remote_dir.strip('/'),
                                                 remote_file))
                    saved_local_fname = os.path.join(data_path, remote_file)
                    req = requests.get(remote_file_path)
                    if req.status_code != 404:
                        open(saved_local_fname, 'wb').write(req.content)
                        i += 1
                    else:
                        logger.info(' '.join(('File not available for',
                                              date.strftime('%d %B %Y'))))
                logger.info('Downloaded {i:} of {n:} files.'.format(i=i, n=n))
            except requests.exceptions.RequestException as exception:
                logger.info(' '.join((exception, '- Files not available for',
                                      date.strftime('%d %B %Y'))))


def list_remote_files(tag, sat_id,
                      remote_site='https://cdaweb.gsfc.nasa.gov',
                      supported_tags=None,
                      user=None, password=None,
                      fake_daily_files_from_monthly=False,
                      two_digit_year_break=None, delimiter=None,
                      year=None, month=None, day=None):
    """Return a Pandas Series of every file for chosen remote data.

    .. deprecated:: 2.2.0
      `year/month/day` keywords will be removed in pysat 3.0.0, they will be
      replaced with a start/stop syntax consistent with the download routine

    This routine is intended to be used by pysat instrument modules supporting
    a particular NASA CDAWeb dataset.

    Parameters
    -----------
    tag : (string or NoneType)
        Denotes type of file to load.  Accepted types are <tag strings>.
        (default=None)
    sat_id : (string or NoneType)
        Specifies the satellite ID for a constellation.
        (default=None)
    remote_site : (string or NoneType)
        Remote site to download data from
        (default='https://cdaweb.gsfc.nasa.gov')
    supported_tags : dict
        dict of dicts. Keys are supported tag names for download. Value is
        a dict with 'dir', 'remote_fname', 'local_fname'. Inteded to be
        pre-set with functools.partial then assigned to new instrument code.
    user : (string or NoneType)
        Username to be passed along to resource with relevant data.
        (default=None)
    password : (string or NoneType)
        User password to be passed along to resource with relevant data.
        (default=None)
    fake_daily_files_from_monthly : bool
        Some CDAWeb instrument data files are stored by month. This flag,
        when true, accomodates this reality with user feedback on a monthly
        time frame.
        (default=False)
    two_digit_year_break : (int or NoneType)
        If filenames only store two digits for the year, then
        '1900' will be added for years >= two_digit_year_break
        and '2000' will be added for years < two_digit_year_break.
        (default=None)
    delimiter : (string or NoneType)
        If filename is delimited, then provide delimiter alone e.g. '_'
        (default=None)
    year : (int or NoneType)
        Selects a given year to return remote files for.  None returns all
        years.
        (default=None)
    month : (int or NoneType)
        Selects a given month to return remote files for.  None returns all
        months.  Requires year to be defined.
        (default=None)
    day : (int or NoneType)
        Selects a given day to return remote files for.  None returns all
        days.  Requires year and month to be defined.
        (default=None)

    Returns
    --------
    pysat.Files.from_os : (pysat._files.Files)
        A class containing the verified available files

    Examples
    --------
    ::

        fname = 'cnofs_vefi_bfield_1sec_{year:04d}{month:02d}{day:02d}_v05.cdf'
        supported_tags = {'dc_b': fname}
        list_remote_files = \
            functools.partial(nasa_cdaweb.list_remote_files,
                              supported_tags=supported_tags)

        fname = 'cnofs_cindi_ivm_500ms_{year:4d}{month:02d}{day:02d}_v01.cdf'
        supported_tags = {'': fname}
        list_remote_files = \
            functools.partial(cdw.list_remote_files,
                              supported_tags=supported_tags)

    """

    import os
    import requests
    import warnings
    from bs4 import BeautifulSoup

    if tag is None:
        tag = ''
    if sat_id is None:
        sat_id = ''
    try:
        inst_dict = supported_tags[sat_id][tag]
    except KeyError:
        raise ValueError('Tag name unknown.')

    # path to relevant file on CDAWeb
    remote_url = remote_site + inst_dict['dir']

    # naming scheme for files on the CDAWeb server
    format_str = inst_dict['remote_fname']

    # Check for appropriate combination of kwargs.  Warn and continue if not.
    if (year is None) and (month is not None):
        warnings.warn("Month keyword requires year.  Ignoring month.")
        month = None
    if ((year is None) or (month is None)) and (day is not None):
        warnings.warn("Day keyword requires year and month.  Ignoring day.")
        day = None

    # Deprecations warnings for changing syntax
    if any([year, month, day]):
        warnings.warn(' '.join(["The year/month/day keywords have been deprecated",
                                "and will be removed in pysat 3.0.0.  Instead,",
                                "use datetime values consistent with the",
                                "start/stop syntax in the download methods."]),
                      DeprecationWarning, stacklevel=2)

    # get a listing of all files
    # determine if we need to walk directories

    # Find Subdirectories and modify remote_url if user input is specified
    dir_split = os.path.split(format_str)
    if len(dir_split[0]) != 0:
        # Get all subdirectories
        subdirs = dir_split[0].split('/')
        # only keep file portion of format
        format_str = dir_split[-1]
        n_layers = len(subdirs)
        # Check for formatted subdirectories if user input is specified
        for subdir in subdirs:
            if (year is not None) and (subdir.find('year') != -1):
                remote_url = '/'.join((remote_url.strip('/'),
                                       subdir.format(year=year)))
                n_layers -= 1
            if (month is not None) and (subdir.find('month') != -1):
                remote_url = '/'.join((remote_url.strip('/'),
                                       subdir.format(month=month)))
                n_layers -= 1
            if (day is not None) and (subdir.find('day') != -1):
                remote_url = '/'.join((remote_url.strip('/'),
                                       subdir.format(day=day)))
                n_layers -= 1

    # Start with file extention as prime target
    targets = ['.' + format_str.split('.')[-1]]

    # Extract file preamble as target - those characters left of variables
    # or wildcards
    fmt_idx = [format_str.find('{')]
    fmt_idx.append(format_str.find('?'))
    fmt_idx.append(format_str.find('*'))

    # Not all characters may exist in a filename.  Remove those that don't.
    fmt_idx = [elem for elem in fmt_idx if elem != -1]

    # If preamble exists, add to targets
    if fmt_idx:
        targets.append(format_str[0:min(fmt_idx)])

    remote_dirs = []
    for level in range(n_layers + 1):
        remote_dirs.append([])
    remote_dirs[0] = ['']

    # Build a list of files using each filename target as a goal
    if n_layers > 1:
        n_loops = 2
        warnings.warn(' '.join(('Current implementation only goes down one',
                                'level of subdirectories.  Try specifying',
                                'a year for more accurate results.')))
    else:
        n_loops = n_layers + 1
    full_files = []

    try:
        for level in range(n_loops):
            for directory in remote_dirs[level]:
                temp_url = '/'.join((remote_url.strip('/'), directory))
                soup = BeautifulSoup(requests.get(temp_url).content, "lxml")
                links = soup.find_all('a', href=True)
                for link in links:
                    # If there is room to go down, look for directories
                    if link['href'].count('/') == 1:
                        remote_dirs[level+1].append(link['href'])
                    else:
                        # If at the endpoint, add matching files to list
                        add_file = True
                        for target in targets:
                            if link['href'].count(target) == 0:
                                add_file = False
                        if add_file:
                            full_files.append(link['href'])
    except requests.exceptions.ConnectionError as merr:
        raise type(merr)(' '.join((str(merr), 'pysat -> Request potentially',
                                   'exceeded the server limit. Please try again',
                                   'using a smaller data range.')))

    # parse remote filenames to get date information
    if delimiter is None:
        stored = pysat._files.parse_fixed_width_filenames(full_files,
                                                          format_str)
    else:
        stored = pysat._files.parse_delimited_filenames(full_files,
                                                        format_str, delimiter)

    # process the parsed filenames and return a properly formatted Series
    stored_list = pysat._files.process_parsed_filenames(stored,
                                                        two_digit_year_break)
    # Downselect to user-specified dates, if needed
    if year is not None:
        mask = (stored_list.index.year == year)
        if month is not None:
            mask = mask & (stored_list.index.month == month)
            if day is not None:
                mask = mask & (stored_list.index.day == day)
        stored_list = stored_list[mask]

    return stored_list



#!/usr/bin/env python
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.1199703
# ----------------------------------------------------------------------------
"""
Coordinate transformation functions for pysat
"""

import datetime as dt
import numpy as np
import pandas as pds

import pysat

logger = pysat.logger


def adjust_cyclic_data(samples, high=2.0 * np.pi, low=0.0):
    """Adjust cyclic values such as longitude to a different scale

    Parameters
    -----------
    samples : array_like
        Input array
    high: float or int
        Upper boundary for circular standard deviation range (default=2 pi)
    low : float or int
        Lower boundary for circular standard deviation range (default=0)
    axis : int or NoneType
        Axis along which standard deviations are computed.  The default is to
        compute the standard deviation of the flattened array

    Returns
    --------
    out_samples : float
        Circular standard deviation

    """

    out_samples = np.asarray(samples)
    sample_range = high - low
    out_samples[out_samples >= high] -= sample_range
    out_samples[out_samples < low] += sample_range

    return out_samples


def update_longitude(inst, lon_name=None, high=180.0, low=-180.0):
    """ Update longitude to the desired range

    Parameters
    ----------
    inst : pysat.Instrument instance
        instrument object to be updated
    lon_name : string
        name of the longtiude data
    high : float
        Highest allowed longitude value (default=180.0)
    low : float
        Lowest allowed longitude value (default=-180.0)

    Returns
    -------
    updates instrument data in column 'lon_name'

    """

    from pysat.utils.coords import adjust_cyclic_data

    if lon_name not in inst.data.keys():
        raise ValueError('unknown longitude variable name')

    new_lon = adjust_cyclic_data(inst[lon_name], high=high, low=low)

    # Update based on data type
    if inst.pandas_format:
        inst[lon_name] = new_lon
    else:
        inst[lon_name].data = new_lon

    return


def calc_solar_local_time(inst, lon_name=None, slt_name='slt',
                          apply_modulus=True, ref_date=None):
    """ Append solar local time to an instrument object

    Parameters
    ----------
    inst : pysat.Instrument instance
        instrument object to be updated
    lon_name : string
        name of the longtiude data key (assumes data are in degrees)
    slt_name : string
        name of the output solar local time data key (default='slt')
    apply_modulus : bool
        If True, SLT values are confined to [0, 24), if False they may be
        positive or negative based on the value of their universal time
        relative to that of the reference date `ref_date`.
        (default=True)
    ref_date : dt.datetime or NoneType
        Reference initial date. If None, will use the date found at
        `inst.date`. Only valid if apply_modulus is True. (default=None)

    Note
    ----
    Updates Instrument data in column specified by slt_name, as well as
    Metadata

    """
    fill_val = np.nan

    if lon_name not in inst.data.keys():
        raise ValueError('unknown longitude variable name')

    if ref_date is not None and apply_modulus:
        istr = 'Keyword `ref_date` only supported if `apply_modulus`=False.'
        logger.info(istr)

    if ref_date is None:
        # Use date information attached to Instrument object
        ref_date = pds.Timestamp(inst.date)
    else:
        # Use user supplied reference date.
        ref_date = pds.Timestamp(ref_date)

    # Convert from numpy epoch nanoseconds to UT seconds of day
    ut_hr = list()
    for nptime in inst.index.values.astype(int):
        # Numpy times come out in nanoseconds and timestamp converts
        # from seconds
        dtime = dt.datetime.utcfromtimestamp(nptime * 1.0e-9)
        ut_hr.append((dtime.hour * 3600.0 + dtime.minute * 60.0
                      + dtime.second + dtime.microsecond * 1.0e-6) / 3600.0)

    ut_hr = np.array(ut_hr)

    # Account for difference in days for calculations without modulus
    if not apply_modulus:
        day_diff = inst.index - ref_date
        day_diff = np.array([diff.days for diff in day_diff])
        ut_hr += day_diff * 24.

    # Calculate solar local time
    if inst[lon_name].shape == ut_hr.shape or inst[lon_name].shape == ():
        if inst[lon_name].shape == ():
            lon = float(inst[lon_name])
        else:
            lon = inst[lon_name]

        slt = ut_hr + lon / 15.0
        coords = inst.index.name
    else:
        # This can only be accessed by xarray input, but longitude may or
        # may not depend on time
        if inst.index.name in inst[lon_name].coords:
            # Initalize the new shape and coordinatesx
            coords = [ckey for ckey in inst[lon_name].dims]
            slt = np.full(shape=inst[lon_name].shape, fill_value=fill_val)

            # Calculate for each UT hr
            for i, hr in enumerate(ut_hr):
                lon = inst[lon_name][inst.index.name == inst.index[0]]
                slt[i] = hr + lon / 15.0
        else:
            # Initialize the new shape and coordinates
            sshape = list(ut_hr.shape)
            sshape.extend(list(inst[lon_name].shape))

            coords = [ckey for ckey in inst[lon_name].dims]
            coords.insert(0, inst.index.name)

            slt = np.full(shape=sshape, fill_value=fill_val)

            # Calculate for each UT hr
            for i, hr in enumerate(ut_hr):
                slt[i] = hr + inst[lon_name] / 15.0

    if apply_modulus:
        # Ensure that solar local time falls between 0 and 24 hours
        slt = np.mod(slt, 24.0)
        min_val = 0.0
        max_val = 24.0
    else:
        # No modulus applied. Values are unbounded.
        min_val = -np.inf
        max_val = np.inf

    # Add the solar local time to the instrument
    if inst.pandas_format:
        inst[slt_name] = pds.Series(slt, index=inst.index)
    else:
        inst.data = inst.data.assign({slt_name: (coords, slt.data)})

    # Add units to the metadata
    inst.meta[slt_name] = {inst.meta.labels.units: 'h',
                           inst.meta.labels.name: "Solar Local Time",
                           inst.meta.labels.desc: "Solar local time in hours",
                           inst.meta.labels.min_val: min_val,
                           inst.meta.labels.max_val: max_val,
                           inst.meta.labels.fill_val: fill_val}

    return



from nose.tools import raises

import pysat
from pysat.instruments.methods import general as gen


class TestGenMethods():

    def setup(self):
        """Runs before every method to create a clean testing setup."""
        fname = 'fake_data_{year:04d}{month:02d}{day:02d}_v05.cdf'
        self.kwargs = {'tag': '', 'sat_id': '', 'data_path': '/fake/path/',
                       'format_str': None,
                       'supported_tags': {'': {'': fname}}}

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.kwargs


class TestRemoveLeadText():
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        # Load a test instrument
        self.testInst = pysat.Instrument('pysat', 'testing', sat_id='12',
                                         clean_level='clean')
        self.testInst.load(2009, 1)
        self.Npts = len(self.testInst['uts'])

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.Npts

    @raises(ValueError)
    def test_remove_prefix_w_bad_target(self):
        self.testInst['ICON_L27_Blurp'] = self.testInst['dummy1']
        gen.remove_leading_text(self.testInst, target=17.5)

    def test_remove_names_wo_target(self):
        self.testInst['ICON_L27_Blurp'] = self.testInst['dummy1']
        gen.remove_leading_text(self.testInst)
        # check variables unchanged
        assert (len(self.testInst['ICON_L27_Blurp']) == self.Npts)
        # check other names untouched
        assert (len(self.testInst['dummy1']) == self.Npts)

    def test_remove_names_w_target(self):
        self.testInst['ICON_L27_Blurp'] = self.testInst['dummy1']
        gen.remove_leading_text(self.testInst, target='ICON_L27')
        # check prepended text removed
        assert len(self.testInst['_Blurp']) == self.Npts
        # check other names untouched
        assert len(self.testInst['dummy1']) == self.Npts
        # check prepended text removed from metadata
        assert '_Blurp' in self.testInst.meta.keys()

    def test_remove_names_w_target_list(self):
        self.testInst['ICON_L27_Blurp'] = self.testInst['dummy1']
        self.testInst['ICON_L23_Bloop'] = self.testInst['dummy1']
        gen.remove_leading_text(self.testInst,
                                target=['ICON_L27', 'ICON_L23_B'])
        # check prepended text removed
        assert len(self.testInst['_Blurp']) == self.Npts
        assert len(self.testInst['loop']) == self.Npts
        # check other names untouched
        assert len(self.testInst['dummy1']) == self.Npts
        # check prepended text removed from metadata
        assert '_Blurp' in self.testInst.meta.keys()
        assert 'loop' in self.testInst.meta.keys()


class TestRemoveLeadTextXarray(TestRemoveLeadText):
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        # Load a test instrument
        self.testInst = pysat.Instrument('pysat', 'testing2d_xarray',
                                         sat_id='12',
                                         clean_level='clean')
        self.testInst.load(2009, 1)
        self.Npts = len(self.testInst['uts'])

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst, self.Npts

    def test_remove_2D_names_w_target(self):
        gen.remove_leading_text(self.testInst, target='variable')
        # check prepended text removed from variables
        assert '_profiles' in self.testInst.data.variables
        assert self.testInst.data['_profiles'].shape[0] == self.Npts
        # check prepended text removed from metadata
        assert '_profiles' in self.testInst.meta.keys()

    def test_remove_2D_names_w_target_list(self):
        gen.remove_leading_text(self.testInst,
                                target=['variable', 'im'])
        # check prepended text removed from variables
        assert '_profiles' in self.testInst.data.variables
        assert self.testInst.data['_profiles'].shape[0] == self.Npts
        assert 'ages' in self.testInst.data.variables
        # check prepended text removed from metadata
        assert '_profiles' in self.testInst.meta.keys()
        assert 'ages' in self.testInst.meta.keys()



# -*- coding: utf-8 -*-
"""Supports OMNI Combined, Definitive, IMF and Plasma Data, and Energetic
Proton Fluxes, Time-Shifted to the Nose of the Earth's Bow Shock, plus Solar
and Magnetic Indices. Downloads data from the NASA Coordinated Data Analysis
Web (CDAWeb). Supports both 5 and 1 minute files.

Properties
----------
platform
    'omni'
name
    'hro'
tag
    Select time between samples, one of {'1min', '5min'}
sat_id
    None supported

Note
----
Files are stored by the first day of each month. When downloading use
omni.download(start, stop, freq='MS') to only download days that could possibly
have data.  'MS' gives a monthly start frequency.

This material is based upon work supported by the
National Science Foundation under Grant Number 1259508.

Any opinions, findings, and conclusions or recommendations expressed in this
material are those of the author(s) and do not necessarily reflect the views
of the National Science Foundation.


Warnings
--------
- Currently no cleaning routine. Though the CDAWEB description indicates that
  these level-2 products are expected to be ok.
- Module not written by OMNI team.


Custom Functions
-----------------
time_shift_to_magnetic_poles
    Shift time from bowshock to intersection with one of the magnetic poles
calculate_clock_angle
    Calculate the clock angle and IMF mag in the YZ plane
calculate_imf_steadiness
    Calculate the IMF steadiness using clock angle and magnitude in the YZ plane
calculate_dayside_reconnection
    Calculate the dayside reconnection rate

"""

from __future__ import print_function
from __future__ import absolute_import

import functools
import logging
import numpy as np
import pandas as pds

import pysat
from pysat.instruments.methods import nasa_cdaweb as cdw
from pysat.instruments.methods import general as mm_gen

logger = logging.getLogger(__name__)

platform = 'omni'
name = 'hro'
tags = {'1min': '1-minute time averaged data',
        '5min': '5-minute time averaged data'}
sat_ids = {'': ['5min']}
_test_dates = {'': {'1min': pysat.datetime(2009, 1, 1),
                    '5min': pysat.datetime(2009, 1, 1)}}

# support list files routine
# use the default CDAWeb method
fname1 = 'omni_hro_1min_{year:4d}{month:02d}{day:02d}_v01.cdf'
fname5 = 'omni_hro_5min_{year:4d}{month:02d}{day:02d}_v01.cdf'
supported_tags = {'': {'1min': fname1,
                       '5min': fname5}}
list_files = functools.partial(mm_gen.list_files,
                               supported_tags=supported_tags,
                               fake_daily_files_from_monthly=True)

# support load routine
# use the default CDAWeb method
load = functools.partial(cdw.load, fake_daily_files_from_monthly=True)

# support download routine
# use the default CDAWeb method
basic_tag1 = {'dir': '/pub/data/omni/omni_cdaweb/hro_1min',
              'remote_fname': '{year:4d}/' + fname1,
              'local_fname': fname1}
basic_tag5 = {'dir': '/pub/data/omni/omni_cdaweb/hro_5min',
              'remote_fname': '{year:4d}/' + fname5,
              'local_fname': fname5}
supported_tags = {'': {'1min': basic_tag1,
                       '5min': basic_tag5}}
download = functools.partial(cdw.download,
                             supported_tags,
                             fake_daily_files_from_monthly=True)
# support listing files currently on CDAWeb
list_remote_files = functools.partial(cdw.list_remote_files,
                                      supported_tags=supported_tags)


def clean(omni):
    for fill_attr in ["fillval", "fill"]:
        # case insensitive check for attribute name
        if omni.meta.has_attr(fill_attr):
            # get real name
            fill_attr = omni.meta.attr_case_name(fill_attr)
            for key in omni.data.columns:
                if key != 'Epoch':
                    idx, = np.where(omni[key] == omni.meta[key, fill_attr])
                    omni[idx, key] = np.nan
    return


def time_shift_to_magnetic_poles(inst):
    """ OMNI data is time-shifted to bow shock. Time shifted again
    to intersections with magnetic pole.

    Parameters
    -----------
    inst : Instrument class object
        Instrument with OMNI HRO data

    Notes
    ---------
    Time shift calculated using distance to bow shock nose (BSN)
    and velocity of solar wind along x-direction.

    Warnings
    --------
    Use at own risk.

    """

    # need to fill in Vx to get an estimate of what is going on
    inst['Vx'] = inst['Vx'].interpolate('nearest')
    inst['Vx'] = inst['Vx'].fillna(method='backfill')
    inst['Vx'] = inst['Vx'].fillna(method='pad')

    inst['BSN_x'] = inst['BSN_x'].interpolate('nearest')
    inst['BSN_x'] = inst['BSN_x'].fillna(method='backfill')
    inst['BSN_x'] = inst['BSN_x'].fillna(method='pad')

    # make sure there are no gaps larger than a minute
    inst.data = inst.data.resample('1T').interpolate('time')

    time_x = inst['BSN_x']*6371.2/-inst['Vx']
    idx, = np.where(np.isnan(time_x))
    if len(idx) > 0:
        logger.info(time_x[idx])
        logger.info(time_x)
    time_x_offset = [pds.DateOffset(seconds=time)
                     for time in time_x.astype(int)]
    new_index = []
    for i, time in enumerate(time_x_offset):
        new_index.append(inst.data.index[i] + time)
    inst.data.index = new_index
    inst.data = inst.data.sort_index()

    return


def calculate_clock_angle(inst):
    """ Calculate IMF clock angle and magnitude of IMF in GSM Y-Z plane

    Parameters
    -----------
    inst : pysat.Instrument
        Instrument with OMNI HRO data

    """

    # Calculate clock angle in degrees
    clock_angle = np.degrees(np.arctan2(inst['BY_GSM'], inst['BZ_GSM']))
    clock_angle[clock_angle < 0.0] += 360.0
    inst['clock_angle'] = pds.Series(clock_angle, index=inst.data.index)

    # Calculate magnitude of IMF in Y-Z plane
    inst['BYZ_GSM'] = pds.Series(np.sqrt(inst['BY_GSM']**2 +
                                         inst['BZ_GSM']**2),
                                 index=inst.data.index)

    return


def calculate_imf_steadiness(inst, steady_window=15, min_window_frac=0.75,
                             max_clock_angle_std=90.0/np.pi, max_bmag_cv=0.5):
    """ Calculate IMF steadiness using clock angle standard deviation and
    the coefficient of variation of the IMF magnitude in the GSM Y-Z plane

    Parameters
    -----------
    inst : pysat.Instrument
        Instrument with OMNI HRO data
    steady_window : int
        Window for calculating running statistical moments in min (default=15)
    min_window_frac : float
        Minimum fraction of points in a window for steadiness to be calculated
        (default=0.75)
    max_clock_angle_std : float
        Maximum standard deviation of the clock angle in degrees (default=22.5)
    max_bmag_cv : float
        Maximum coefficient of variation of the IMF magnitude in the GSM
        Y-Z plane (default=0.5)

    """

    from pysat.utils import stats as pystats

    # We are not going to interpolate through missing values
    sample_rate = int(inst.tag[0])
    max_wnum = np.floor(steady_window / sample_rate)
    if max_wnum != steady_window / sample_rate:
        steady_window = max_wnum * sample_rate
        logger.warning("sample rate is not a factor of the statistical window")
        logger.warning("new statistical window is {:.1f}".format(steady_window))

    min_wnum = int(np.ceil(max_wnum * min_window_frac))

    # Calculate the running coefficient of variation of the BYZ magnitude
    byz_mean = inst['BYZ_GSM'].rolling(min_periods=min_wnum, center=True,
                                       window=steady_window).mean()
    byz_std = inst['BYZ_GSM'].rolling(min_periods=min_wnum, center=True,
                                      window=steady_window).std()
    inst['BYZ_CV'] = pds.Series(byz_std / byz_mean, index=inst.data.index)

    # Calculate the running circular standard deviation of the clock angle
    circ_kwargs = {'high': 360.0, 'low': 0.0}

    ca_std = \
        inst['clock_angle'].rolling(min_periods=min_wnum,
                                    window=steady_window,
                                    center=True).apply(pystats.nan_circstd,
                                                       kwargs=circ_kwargs,
                                                       raw=True)
    inst['clock_angle_std'] = pds.Series(ca_std, index=inst.data.index)

    # Determine how long the clock angle and IMF magnitude are steady
    imf_steady = np.zeros(shape=inst.data.index.shape)

    steady = False
    for i, cv in enumerate(inst.data['BYZ_CV']):
        if steady:
            del_min = int((inst.data.index[i] -
                           inst.data.index[i-1]).total_seconds() / 60.0)
            if np.isnan(cv) or np.isnan(ca_std[i]) or del_min > sample_rate:
                # Reset the steadiness flag if fill values are encountered, or
                # if an entry is missing
                steady = False

        if cv <= max_bmag_cv and ca_std[i] <= max_clock_angle_std:
            # Steadiness conditions have been met
            if steady:
                imf_steady[i] = imf_steady[i-1]

            imf_steady[i] += sample_rate
            steady = True

    inst['IMF_Steady'] = pds.Series(imf_steady, index=inst.data.index)
    return


def calculate_dayside_reconnection(inst):
    """ Calculate the dayside reconnection rate (Milan et al. 2014)

    Parameters
    -----------
    inst : pysat.Instrument
        Instrument with OMNI HRO data, requires BYZ_GSM and clock_angle

    Notes
    --------
    recon_day = 3.8 Re (Vx / 4e5 m/s)^1/3 Vx B_yz (sin(theta/2))^9/2

    """

    rearth = 6371008.8
    sin_htheta = np.power(np.sin(np.radians(0.5 * inst['clock_angle'])), 4.5)
    byz = inst['BYZ_GSM'] * 1.0e-9
    vx = inst['flow_speed'] * 1000.0

    recon_day = 3.8 * rearth * vx * byz * sin_htheta * np.power((vx / 4.0e5),
                                                                1.0/3.0)
    inst['recon_day'] = pds.Series(recon_day, index=inst.data.index)
    return



"""
Creates a constellation with 5 testing instruments

Attributes
----------
instruments : list
    List of pysat.Instrument objects

"""
import pysat

instruments = [pysat.Instrument('pysat', 'testing', clean_level='clean')
               for i in range(5)]



from __future__ import print_function
from __future__ import absolute_import

import os
import warnings
import numpy as np
import pandas as pds
# python 2/3 compatibility
try:
    basestring
except NameError:
    basestring = str

from pysat import DataFrame, Series


class Meta(object):
    """ Stores metadata for Instrument instance, similar to CF-1.6 netCDFdata
    standard.

    Parameters
    ----------
    metadata : pandas.DataFrame
        DataFrame should be indexed by variable name that contains at minimum
        the standard_name (name), units, and long_name for the data stored in
        the associated pysat Instrument object.

    units_label : str
        String used to label units in storage. Defaults to 'units'.
    name_label : str
        String used to label long_name in storage. Defaults to 'long_name'.
    notes_label : str
        String used to label 'notes' in storage. Defaults to 'notes'
    desc_label : str
        String used to label variable descriptions in storage.
        Defaults to 'desc'
    plot_label : str
        String used to label variables in plots. Defaults to 'label'
    axis_label : str
        Label used for axis on a plot. Defaults to 'axis'
    scale_label : str
        string used to label plot scaling type in storage. Defaults to 'scale'
    min_label : str
        String used to label typical variable value min limit in storage.
        Defaults to 'value_min'
    max_label : str
        String used to label typical variable value max limit in storage.
        Defaults to 'value_max'
    fill_label : str
        String used to label fill value in storage. Defaults to 'fill' per
        netCDF4 standard


    Attributes
    ----------
    data : pandas.DataFrame
        index is variable standard name, 'units', 'long_name', and other
        defaults are also stored along with additional user provided labels.

    units_label : str
        String used to label units in storage. Defaults to 'units'.
    name_label : str
        String used to label long_name in storage. Defaults to 'long_name'.
    notes_label : str
       String used to label 'notes' in storage. Defaults to 'notes'
    desc_label : str
       String used to label variable descriptions in storage.
       Defaults to 'desc'
    plot_label : str
       String used to label variables in plots. Defaults to 'label'
    axis_label : str
        Label used for axis on a plot. Defaults to 'axis'
    scale_label : str
       string used to label plot scaling type in storage. Defaults to 'scale'
    min_label : str
       String used to label typical variable value min limit in storage.
       Defaults to 'value_min'
    max_label : str
       String used to label typical variable value max limit in storage.
       Defaults to 'value_max'
    fill_label : str
        String used to label fill value in storage. Defaults to 'fill' per
        netCDF4 standard
    export_nan: list
        List of labels that should be exported even if their value is nan.
        By default, metadata with a value of nan will be exluded from export.


    Notes
    -----
    Meta object preserves the case of variables and attributes as it first
    receives the data. Subsequent calls to set new metadata with the same
    variable or attribute will use case of first call. Accessing or setting
    data thereafter is case insensitive. In practice, use is case insensitive
    but the original case is preserved. Case preseveration is built in to
    support writing files with a desired case to meet standards.

    Metadata for higher order data objects, those that have
    multiple products under a single variable name in a pysat.Instrument
    object, are stored by providing a Meta object under the single name.

    Supports any custom metadata values in addition to the expected metadata
    attributes (units, name, notes, desc, plot_label, axis, scale, value_min,
    value_max, and fill). These base attributes may be used to programatically
    access and set types of metadata regardless of the string values used for
    the attribute. String values for attributes may need to be changed
    depending upon the standards of code or files interacting with pysat.

    Meta objects returned as part of pysat loading routines are automatically
    updated to use the same values of plot_label, units_label, etc. as found
    on the pysat.Instrument object.

    Examples
    --------
    ::

        # instantiate Meta object, default values for attribute labels are used
        meta = pysat.Meta()
        # set a couple base units
        # note that other base parameters not set below will
        # be assigned a default value
        meta['name'] = {'long_name':string, 'units':string}
        # update 'units' to new value
        meta['name'] = {'units':string}
        # update 'long_name' to new value
        meta['name'] = {'long_name':string}
        # attach new info with partial information, 'long_name' set to 'name2'
        meta['name2'] = {'units':string}
        # units are set to '' by default
        meta['name3'] = {'long_name':string}

        # assigning custom meta parameters
        meta['name4'] = {'units':string, 'long_name':string
                         'custom1':string, 'custom2':value}
        meta['name5'] = {'custom1':string, 'custom3':value}

        # assign multiple variables at once
        meta[['name1', 'name2']] = {'long_name':[string1, string2],
                                    'units':[string1, string2],
                                    'custom10':[string1, string2]}

        # assiging metadata for n-Dimensional variables
        meta2 = pysat.Meta()
        meta2['name41'] = {'long_name':string, 'units':string}
        meta2['name42'] = {'long_name':string, 'units':string}
        meta['name4'] = {'meta':meta2}
        # or
        meta['name4'] = meta2
        meta['name4'].children['name41']

        # mixture of 1D and higher dimensional data
        meta = pysat.Meta()
        meta['dm'] = {'units':'hey', 'long_name':'boo'}
        meta['rpa'] = {'units':'crazy', 'long_name':'boo_whoo'}
        meta2 = pysat.Meta()
        meta2[['higher', 'lower']] = {'meta':[meta, None],
                                      'units':[None, 'boo'],
                                      'long_name':[None, 'boohoo']}

        # assign from another Meta object
        meta[key1] = meta2[key2]

        # access fill info for a variable, presuming default label
        meta[key1, 'fill']
        # access same info, even if 'fill' not used to label fill values
        meta[key1, meta.fill_label]


        # change a label used by Meta object
        # note that all instances of fill_label
        # within the meta object are updated
        meta.fill_label = '_FillValue'
        meta.plot_label = 'Special Plot Variable'
        # this feature is useful when converting metadata within pysat
        # so that it is consistent with externally imposed file standards

    """

    def __init__(self, metadata=None, units_label='units',
                 name_label='long_name', notes_label='notes',
                 desc_label='desc', plot_label='label', axis_label='axis',
                 scale_label='scale', min_label='value_min',
                 max_label='value_max', fill_label='fill',
                 export_nan=[]):

        # set mutability of Meta attributes
        self.mutable = True

        # set units and name labels directly
        self._units_label = units_label
        self._name_label = name_label
        self._notes_label = notes_label
        self._desc_label = desc_label
        self._plot_label = plot_label
        self._axis_label = axis_label
        self._scale_label = scale_label
        self._min_label = min_label
        self._max_label = max_label
        self._fill_label = fill_label
        # by default metadata with a value of nan will not be exported
        # unless the name is in the _export_nan list. Initialize the list
        # with the fill label, since it is reasonable to assume that a fill
        # value of nan would be intended to be exported
        self._export_nan = [fill_label] + export_nan
        # init higher order (nD) data structure container, a dict
        self._ho_data = {}
        # use any user provided data to instantiate object with data
        # attirube unit and name labels are called within
        if metadata is not None:
            if isinstance(metadata, DataFrame):
                self._data = metadata
                # make sure defaults are taken care of for required metadata
                self.accept_default_labels(self)
            else:
                raise ValueError(''.join(('Input must be a pandas DataFrame',
                                          'type. See other constructors for',
                                          ' alternate inputs.')))
        else:
            self._data = DataFrame(None, columns=[self._units_label,
                                                  self._name_label,
                                                  self._desc_label,
                                                  self._plot_label,
                                                  self._axis_label,
                                                  self._scale_label,
                                                  self.notes_label,
                                                  self._min_label,
                                                  self._max_label,
                                                  self._fill_label])

        # establish attributes intrinsic to object, before user can
        # add any
        self._base_attr = dir(self)

    @property
    def ho_data(self):
        return self._ho_data

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, new_frame):
        self._data = new_frame
        # self.keys = self._data.columns.lower()

    @ho_data.setter
    def ho_data(self, new_dict):
        self._ho_data = new_dict

    @property
    def empty(self):
        """Return boolean True if there is no metadata"""

        # only need to check on lower data since lower data
        # is set when higher metadata assigned
        if self.data.empty:
            return True
        else:
            return False

    def merge(self, other):
        """Adds metadata variables to self that are in other but not in self.

        Parameters
        ----------
        other : pysat.Meta

        """

        for key in other.keys():
            if key not in self:
                # copies over both lower and higher dimensional data
                self[key] = other[key]

    def drop(self, names):
        """Drops variables (names) from metadata."""

        # drop lower dimension data
        self.data = self._data.drop(names, axis=0)
        # drop higher dimension data
        for name in names:
            if name in self._ho_data:
                _ = self._ho_data.pop(name)

    def keep(self, keep_names):
        """Keeps variables (keep_names) while dropping other parameters

        Parameters
        ----------
        keep_names : list-like
            variables to keep
        """
        keep_names = [self.var_case_name(name) for name in keep_names]
        current_names = self._data.index
        drop_names = []
        for name in current_names:
            if name not in keep_names:
                drop_names.append(name)
        self.drop(drop_names)

    def apply_default_labels(self, other):
        """Applies labels for default meta labels from self onto other.

        Parameters
        ----------
        other : Meta
            Meta object to have default labels applied

        Returns
        -------
        Meta

        """
        other_updated = other.copy()
        other_updated.units_label = self.units_label
        other_updated.name_label = self.name_label
        other_updated.notes_label = self.notes_label
        other_updated.desc_label = self.desc_label
        other_updated.plot_label = self.plot_label
        other_updated.axis_label = self.axis_label
        other_updated.scale_label = self.scale_label
        other_updated.min_label = self.min_label
        other_updated.max_label = self.max_label
        other_updated.fill_label = self.fill_label
        return other_updated

    def accept_default_labels(self, other):
        """Applies labels for default meta labels from other onto self.

        Parameters
        ----------
        other : Meta
            Meta object to take default labels from

        Returns
        -------
        Meta

        """

        self.units_label = other.units_label
        self.name_label = other.name_label
        self.notes_label = other.notes_label
        self.desc_label = other.desc_label
        self.plot_label = other.plot_label
        self.axis_label = other.axis_label
        self.scale_label = other.scale_label
        self.min_label = other.min_label
        self.max_label = other.max_label
        self.fill_label = other.fill_label
        return

    def __contains__(self, other):
        """case insensitive check for variable name"""

        if other.lower() in [i.lower() for i in self.keys()]:
            return True
        if other.lower() in [i.lower() for i in self.keys_nD()]:
            return True
        return False

    def __repr__(self):
        return 'pysat.MetaData'

    def __str__(self, recurse=True):
        """String describing Meta instance, variables, and attributes"""

        # cover 1D parameters
        if recurse:
            output_str = 'Metadata for 1D variables\n'
        else:
            output_str = ''

        for ind in self.keys():
            output_str += ind.ljust(30)
        output_str += '\n\n'
        output_str += 'Tracking the following:\n'
        for col in self.attrs():
            output_str += col.ljust(30)

        output_str += '\n'
        if recurse:
            for item_name in self.keys_nD():
                output_str += '\n\n'
                output_str += 'Metadata for '+item_name+'\n'
                output_str += self.ho_data[item_name].__str__(False)

        return output_str

    def _insert_default_values(self, input_name):

        default_str = ''
        default_nan = np.NaN
        labels = [self.units_label, self.name_label, self.notes_label,
                  self.desc_label, self.plot_label, self.axis_label,
                  self.scale_label, self.min_label, self.max_label,
                  self.fill_label]
        defaults = [default_str, input_name, default_str, default_str,
                    input_name, input_name, 'linear', default_nan,
                    default_nan, default_nan]
        self._data.loc[input_name, labels] = defaults


    def __setattr__(self, name, value):
        """Conditionally sets attributes based on self.mutable flag
        @properties are assumed to be mutable.
        We avoid recursively setting properties using
        method from https://stackoverflow.com/a/15751135
        """

        # mutable handled explicitly to avoid recursion
        if name != 'mutable':

            # check if this attribute is a property
            propobj = getattr(self.__class__, name, None)
            if isinstance(propobj, property):
                # check if the property is settable
                if propobj.fset is None:
                    raise AttributeError(''.join("can't set attribute - ",
                                        "property has no fset"))

                # make mutable in case fset needs it to be
                mutable_tmp = self.mutable
                self.mutable = True

                # set the property
                propobj.fset(self, value)

                # restore mutability flag
                self.mutable = mutable_tmp
            else:
                # a normal attribute
                if self.mutable:
                    # use Object to avoid recursion
                    super(Meta, self).__setattr__(name, value)
                else:
                    raise AttributeError(''.join(("cannot set attribute - ",
                                                    "object's attributes are immutable")))
        else:
            super(Meta, self).__setattr__(name, value)


    def __setitem__(self, names, input_data):
        """Convenience method for adding metadata."""

        if isinstance(input_data, dict):
            # if not passed an iterable, make it one
            if isinstance(names, basestring):
                names = [names]
                for key in input_data:
                    input_data[key] = [input_data[key]]
            elif isinstance(names, slice) and (names.step is None):
                # Check for instrument[indx,:] or instrument[idx] usage
                names = list(self.data.keys())
            # make sure the variable names are in good shape
            # Meta object is case insensitive but case preserving
            # convert given names into ones Meta has already seen
            # if new, then input names become the standard
            names = [self.var_case_name(name) for name in names]
            for name in names:
                if name not in self:
                    self._insert_default_values(name)
            # check if input dict empty
            if input_data.keys() == []:
                # meta wasn't actually assigned by user, empty call
                # we can head out - we've assigned defaults if first data
                return
            # perform some checks on the data
            # make sure number of inputs matches number of metadata inputs
            for key in input_data:
                if len(names) != len(input_data[key]):
                    raise ValueError(''.join(('Length of names and inputs',
                                              ' must be equal.')))
            # make sure the attribute names are in good shape
            # check name of attributes against existing attribute names
            # if attribute name exists somewhere, then case of existing
            # attribute
            # will be enforced upon new data by default for consistency
            keys = [i for i in input_data]
            for name in keys:
                new_name = self.attr_case_name(name)
                if new_name != name:
                    input_data[new_name] = input_data.pop(name)

            # time to actually add the metadata
            for key in input_data:
                if key not in ['children', 'meta']:
                    for i, name in enumerate(names):
                        to_be_set = input_data[key][i]
                        if hasattr(to_be_set, '__iter__') and \
                                not isinstance(to_be_set, basestring):
                            # we have some list-like object
                            # can only store a single element
                            if len(to_be_set) == 0:
                                # empty list, ensure there is something
                                to_be_set = ['']
                            if isinstance(to_be_set[0], basestring):
                                self._data.loc[name, key] = \
                                    '\n\n'.join(to_be_set)
                            else:
                                warnings.warn(' '.join(('Array elements are',
                                                        'not allowed in meta.',
                                                        'Dropping input :',
                                                        key)))
                        else:
                            self._data.loc[name, key] = to_be_set
                else:
                    # key is 'meta' or 'children'
                    # process higher order stuff. Meta inputs could be part of
                    # larger multiple parameter assignment
                    # so not all names may actually have 'meta' to add
                    for j, (item, val) in enumerate(zip(names,
                                                        input_data['meta'])):
                        if val is not None:
                            # assign meta data, recursive call....
                            # heads to if Meta instance call
                            self[item] = val

        elif isinstance(input_data, Series):
            # outputs from Meta object are a Series.
            # thus this takes in input from a Meta object
            # set data usind standard assignment via a dict
            in_dict = input_data.to_dict()
            if 'children' in in_dict:
                child = in_dict.pop('children')
                if child is not None:
                    # if not child.data.empty:
                    self.ho_data[names] = child
            # remaining items are simply assigned
            self[names] = in_dict

        elif isinstance(input_data, Meta):
            # dealing with higher order data set
            # names is only a single name here (by choice for support)
            if (names in self._ho_data) and (input_data.empty):
                # no actual metadata provided and there is already some
                # higher order metadata in self
                return

            # get Meta approved variable names
            new_item_name = self.var_case_name(names)
            # ensure that Meta labels of object to be assigned
            # are consistent with self
            # input_data accepts self's labels
            input_data.accept_default_labels(self)

            # go through and ensure Meta object to be added has variable and
            # attribute names consistent with other variables and attributes
            # this covers custom attributes not handled by default routine
            # above
            attr_names = input_data.attrs()
            new_names = []
            for name in attr_names:
                new_names.append(self.attr_case_name(name))
            input_data.data.columns = new_names
            # same thing for variables
            var_names = input_data.data.index
            new_names = []
            for name in var_names:
                new_names.append(self.var_case_name(name))
            input_data.data.index = new_names
            # assign Meta object now that things are consistent with Meta
            # object settings
            # but first, make sure there are lower dimension metadata
            # parameters, passing in an empty dict fills in defaults
            # if there is no existing metadata info
            self[new_item_name] = {}
            # now add to higher order data
            self._ho_data[new_item_name] = input_data

    def __getitem__(self, key):
        """Convenience method for obtaining metadata.

        Maps to pandas DataFrame.loc method.

        Examples
        --------
        ::

            meta['name']

            meta[ 'name1', 'units' ]

            meta[[ 'name1', 'name2'], 'units']

            meta[:, 'units']

            for higher order data

            meta[ 'name1', 'subvar', 'units' ]

            meta[ 'name1', ('units', 'scale') ]

        """
        # if key is a tuple, looking at index, column access pattern

        def match_name(func, name, names):
            """Applies func on name(s) depending on name type"""
            if isinstance(name, basestring):
                return func(name)
            elif isinstance(name, slice):
                return [func(nn) for nn in names[name]]
            else:
                # assume iterable
                return [func(nn) for nn in name]

        if isinstance(key, tuple):
            # if tuple length is 2, index, column
            if len(key) == 2:
                new_index = match_name(self.var_case_name, key[0],
                                        self.data.index)
                new_name = match_name(self.attr_case_name, key[1],
                                        self.data.columns)
                return self.data.loc[new_index, new_name]

            # if tuple length is 3, index, child_index, column
            elif len(key) == 3:
                new_index = self.var_case_name(key[0])
                new_child_index = self.var_case_name(key[1])
                new_name = self.attr_case_name(key[2])
                return self.ho_data[new_index].data.loc[new_child_index,
                                                        new_name]

        elif isinstance(key, list):
            return self[key, :]

        elif isinstance(key, basestring):
            # ensure variable is present somewhere
            if key in self:
                # get case preserved string for variable name
                new_key = self.var_case_name(key)
                # if new_key in self.keys():
                # don't need to check if in lower, all variables
                # are always in the lower metadata
                meta_row = self.data.loc[new_key]
                if new_key in self.keys_nD():
                    meta_row.at['children'] = self.ho_data[new_key].copy()
                else:
                    # empty_meta = Meta()
                    # self.apply_default_labels(empty_meta)
                    # Following line issues a pandas SettingWithCopyWarning
                    meta_row.at['children'] = None  # empty_meta
                return meta_row
                # else:
                #     return pds.Series([self.ho_data[new_key].copy()],
                #                       index=['children'])
            else:
                raise KeyError('Key not found in MetaData')
        else:
            raise NotImplementedError("No way to handle MetaData key {}".format(
                key.__repr__()))

    def _label_setter(self, new_label, current_label, attr_label,
                      default=np.NaN, use_names_default=False):
        """Generalized setter of default meta attributes

        Parameters
        ----------
        new_label : str
            New label to use in the Meta object
        current_label : str
            The hidden attribute to be updated that actually stores metadata
        default :
            Deafult setting to use for label if there is no attribute
            value
        use_names_default : bool
            if True, MetaData variable names are used as the default
            value for the specified Meta attributes settings

        Examples
        --------
        :
                @name_label.setter
                def name_label(self, new_label):
                    self._label_setter(new_label, self._name_label,
                                        use_names_default=True)

        Notes
        -----
        Not intended for end user

        """

        if new_label not in self.attrs():
            # new label not in metadata, including case
            # update existing label, if present
            if current_label in self.attrs():
                # old label exists and has expected case
                self.data.loc[:, new_label] = self.data.loc[:, current_label]
                self.data.drop(current_label, axis=1, inplace=True)
            else:
                if self.has_attr(current_label):
                    # there is something like label, wrong case though
                    current_label = self.attr_case_name(current_label)
                    self.data.loc[:, new_label] = \
                        self.data.loc[:, current_label]
                    self.data.drop(current_label, axis=1, inplace=True)
                else:
                    # there is no existing label
                    # setting for the first time
                    if use_names_default:
                        self.data[new_label] = self.data.index
                    else:
                        self.data[new_label] = default
            # check higher order structures as well
            # recursively change labels here
            for key in self.keys_nD():
                setattr(self.ho_data[key], attr_label, new_label)

        # now update 'hidden' attribute value
        # current_label = new_label
        setattr(self, ''.join(('_', attr_label)), new_label)

    @property
    def units_label(self):
        return self._units_label

    @property
    def name_label(self):
        return self._name_label

    @property
    def notes_label(self):
        return self._notes_label

    @property
    def desc_label(self):
        return self._desc_label

    @property
    def plot_label(self):
        return self._plot_label

    @property
    def axis_label(self):
        return self._axis_label

    @property
    def scale_label(self):
        return self._scale_label

    @property
    def min_label(self):
        return self._min_label

    @property
    def max_label(self):
        return self._max_label

    @property
    def fill_label(self):
        return self._fill_label

    @units_label.setter
    def units_label(self, new_label):
        self._label_setter(new_label, self._units_label, 'units_label', '')

    @name_label.setter
    def name_label(self, new_label):
        self._label_setter(new_label, self._name_label, 'name_label',
                           use_names_default=True)

    @notes_label.setter
    def notes_label(self, new_label):
        self._label_setter(new_label, self._notes_label, 'notes_label', '')

    @desc_label.setter
    def desc_label(self, new_label):
        self._label_setter(new_label, self._desc_label, 'desc_label', '')

    @plot_label.setter
    def plot_label(self, new_label):
        self._label_setter(new_label, self._plot_label, 'plot_label',
                           use_names_default=True)

    @axis_label.setter
    def axis_label(self, new_label):
        self._label_setter(new_label, self._axis_label, 'axis_label',
                           use_names_default=True)

    @scale_label.setter
    def scale_label(self, new_label):
        self._label_setter(new_label, self._scale_label, 'scale_label',
                           'linear')

    @min_label.setter
    def min_label(self, new_label):
        self._label_setter(new_label, self._min_label, 'min_label', np.NaN)

    @max_label.setter
    def max_label(self, new_label):
        self._label_setter(new_label, self._max_label, 'max_label', np.NaN)

    @fill_label.setter
    def fill_label(self, new_label):
        self._label_setter(new_label, self._fill_label, 'fill_label', np.NaN)

    def var_case_name(self, name):
        """Provides stored name (case preserved) for case insensitive input

        If name is not found (case-insensitive check) then name is returned,
        as input. This function is intended to be used to help ensure the
        case of a given variable name is the same across the Meta object.

        Parameters
        ----------
        name : str
            variable name in any case

        Returns
        -------
        str
            string with case preserved as in metaobject

        """

        lower_name = name.lower()
        if name in self:
            for i in self.keys():
                if lower_name == i.lower():
                    return i
            for i in self.keys_nD():
                if lower_name == i.lower():
                    return i
        return name

    def keys(self):
        """Yields variable names stored for 1D variables"""

        for i in self.data.index:
            yield i

    def keys_nD(self):
        """Yields keys for higher order metadata"""

        for i in self.ho_data:
            yield i

    def attrs(self):
        """Yields metadata products stored for each variable name"""

        for i in self.data.columns:
            yield i

    def has_attr(self, name):
        """Returns boolean indicating presence of given attribute name

        Case-insensitive check

        Notes
        -----
        Does not check higher order meta objects

        Parameters
        ----------
        name : str
            name of variable to get stored case form

        Returns
        -------
        bool
            True if case-insesitive check for attribute name is True

        """

        if name.lower() in [i.lower() for i in self.data.columns]:
            return True
        return False

    def attr_case_name(self, name):
        """Returns preserved case name for case insensitive value of name.

        Checks first within standard attributes. If not found there, checks
        attributes for higher order data structures. If not found, returns
        supplied name as it is available for use. Intended to be used to help
        ensure that the same case is applied to all repetitions of a given
        variable name.

        Parameters
        ----------
        name : str
            name of variable to get stored case form

        Returns
        -------
        str
            name in proper case
        """

        lower_name = name.lower()
        for i in self.attrs():
            if lower_name == i.lower():
                return i
        # check if attribute present in higher order structures
        for key in self.keys_nD():
            for i in self[key].children.attrs():
                if lower_name == i.lower():
                    return i
        # nothing was found if still here
        # pass name back, free to be whatever
        return name

    def concat(self, other, strict=False):
        """Concats two metadata objects together.

        Parameters
        ----------
        other : Meta
            Meta object to be concatenated
        strict : bool
            if True, ensure there are no duplicate variable names

        Notes
        -----
        Uses units and name label of self if other is different

        Returns
        -------
        Meta
            Concatenated object
        """

        mdata = self.copy()
        # checks
        if strict:
            for key in other.keys():
                if key in mdata:
                    raise RuntimeError(''.join(('Duplicated keys (variable ',
                                                'names) across Meta ',
                                                'objects in keys().')))
            for key in other.keys_nD():
                if key in mdata:

                    raise RuntimeError(''.join(('Duplicated keys (variable ',
                                                'names) across Meta '
                                                'objects in keys_nD().')))

        # make sure labels between the two objects are the same
        other_updated = self.apply_default_labels(other)
        # concat 1D metadata in data frames to copy of
        # current metadata
        for key in other_updated.keys():
            mdata.data.loc[key] = other.data.loc[key]
        # add together higher order data
        for key in other_updated.keys_nD():
            mdata.ho_data[key] = other.ho_data[key]

        return mdata

    def copy(self):
        from copy import deepcopy as deepcopy
        """Deep copy of the meta object."""
        return deepcopy(self)

    def pop(self, name):
        """Remove and return metadata about variable

        Parameters
        ----------
        name : str
            variable name

        Returns
        -------
        pandas.Series
            Series of metadata for variable
        """
        # check if present
        if name in self:
            # get case preserved name for variable
            new_name = self.var_case_name(name)
            # check if 1D or nD
            if new_name in self.keys():
                output = self[new_name]
                self.data.drop(new_name, inplace=True, axis=0)
            else:
                output = self.ho_data.pop(new_name)

            return output
        else:
            raise KeyError('Key not present in metadata variables')

    def transfer_attributes_to_instrument(self, inst, strict_names=False):
        """Transfer non-standard attributes in Meta to Instrument object.

        Pysat's load_netCDF and similar routines are only able to attach
        netCDF4 attributes to a Meta object. This routine identifies these
        attributes and removes them from the Meta object. Intent is to
        support simple transfers to the pysat.Instrument object.

        Will not transfer names that conflict with pysat default attributes.

        Parameters
        ----------
        inst : pysat.Instrument
            Instrument object to transfer attributes to
        strict_names : boolean (False)
            If True, produces an error if the Instrument object already
            has an attribute with the same name to be copied.

        Returns
        -------
        None
            pysat.Instrument object modified in place with new attributes
        """

        # base Instrument attributes
        banned = inst._base_attr
        # get base attribute set, and attributes attached to instance
        base_attrb = self._base_attr
        this_attrb = dir(self)
        # collect these attributes into a dict
        adict = {}
        transfer_key = []
        for key in this_attrb:
            if key not in banned:
                if key not in base_attrb:
                    # don't store _ leading attributes
                    if key[0] != '_':
                        adict[key] = self.__getattribute__(key)
                        transfer_key.append(key)

        # store any non-standard attributes in Instrument
        # get list of instrument objects attributes first
        # to check if a duplicate

        # instrument attributes are now inst.meta attributes
        inst_attr = dir(inst)

        for key in transfer_key:
            if key not in banned:
                if key not in inst_attr:
                    inst.__setattr__(key, adict[key])
                else:
                    if not strict_names:
                        # new_name = 'pysat_attr_'+key
                        inst.__setattr__(key, adict[key])
                    else:
                        raise RuntimeError(''.join(('Attribute ', key,
                                                    ' attached to Meta object',
                                                    ' can not be transferred',
                                                    ' as it already exists in',
                                                    ' the Instrument object.'
                                                    )))
        # return inst

    def __eq__(self, other):
        """
        Check equality between Meta instances. Good for testing.

        Checks if variable names, attribute names, and metadata values
        are all equal between to Meta objects. Note that this comparison
        treats np.NaN == np.NaN as True.

        Name comparison is case-sensitive.

        """

        if isinstance(other, Meta):
            # check first if variables and attributes are the same
            # quick check on length
            keys1 = [i for i in self.keys()]
            keys2 = [i for i in other.keys()]
            if len(keys1) != len(keys2):
                return False
            # now iterate over each of the keys in the first one
            # don't need to iterate over second one, if all of the first
            # in the second we are good. No more or less items in second from
            # check earlier.
            for key in keys1:
                if key not in keys2:
                    return False
            # do same checks on attributes
            attrs1 = [i for i in self.attrs()]
            attrs2 = [i for i in other.attrs()]
            if len(attrs1) != len(attrs2):
                return False
            for attr in attrs1:
                if attr not in attrs2:
                    return False
            # now check the values of all elements now that we know all
            # variable and attribute names are the same
            for key in self.keys():
                for attr in self.attrs():
                    if not (self[key, attr] == other[key, attr]):
                        # np.nan is not equal to anything
                        # if both values are NaN, ok in my book
                        try:
                            if not (np.isnan(self[key, attr]) and
                                    np.isnan(other[key, attr])):
                                # one or both are not NaN and they aren't equal
                                # test failed
                                return False
                        except TypeError:
                            # comparison above gets unhappy with string inputs
                            return False

            # check through higher order products
            # in the same manner as code above
            keys1 = [i for i in self.keys_nD()]
            keys2 = [i for i in other.keys_nD()]
            if len(keys1) != len(keys2):
                return False
            for key in keys1:
                if key not in keys2:
                    return False
            # do same check on all sub variables within each nD key
            for key in self.keys_nD():
                keys1 = [i for i in self[key].children.keys()]
                keys2 = [i for i in other[key].children.keys()]
                if len(keys1) != len(keys2):
                    return False
                for key_check in keys1:
                    if key_check not in keys2:
                        return False
                # check if attributes are the same
                attrs1 = [i for i in self[key].children.attrs()]
                attrs2 = [i for i in other[key].children.attrs()]
                if len(attrs1) != len(attrs2):
                    return False
                for attr in attrs1:
                    if attr not in attrs2:
                        return False
                # now time to check if all elements are individually equal
                for key2 in self[key].children.keys():
                    for attr in self[key].children.attrs():
                        if not (self[key].children[key2, attr] ==
                                other[key].children[key2, attr]):
                            try:
                                if not (np.isnan(self[key].children[key2,
                                                                    attr]) and
                                        np.isnan(other[key].children[key2,
                                                                     attr])):
                                    return False
                            except TypeError:
                                # comparison above gets unhappy with string
                                # inputs
                                return False
            # if we made it this far, things are good
            return True
        else:
            # wasn't even the correct class
            return False

    @classmethod
    def from_csv(cls, name=None, col_names=None, sep=None, **kwargs):
        """Create instrument metadata object from csv.

        Parameters
        ----------
        name : string
            absolute filename for csv file or name of file
            stored in pandas instruments location
        col_names : list-like collection of strings
            column names in csv and resultant meta object
        sep : string
            column seperator for supplied csv filename

        Note
        ----
        column names must include at least ['name', 'long_name', 'units'],
        assumed if col_names is None.
        """
        import pysat
        req_names = ['name', 'long_name', 'units']
        if col_names is None:
            col_names = req_names
        elif not all([i in col_names for i in req_names]):
            raise ValueError('col_names must include name, long_name, units.')

        if sep is None:
            sep = ','

        if name is None:
            raise ValueError('Must supply an instrument name or file path.')
        elif not isinstance(name, str):
            raise ValueError('keyword name must be related to a string')
        elif not os.path.isfile(name):
            # Not a real file, assume input is a pysat instrument name
            # and look in the standard pysat location.
            test = os.path.join(pysat.__path__[0], 'instruments', name)
            if os.path.isfile(test):
                name = test
            else:
                # trying to form an absolute path for success
                test = os.path.abspath(name)
                if not os.path.isfile(test):
                    raise ValueError("Unable to create valid file path.")
                else:
                    # success
                    name = test

        mdata = pds.read_csv(name, names=col_names, sep=sep, **kwargs)

        if not mdata.empty:
            # make sure the data name is the index
            mdata.index = mdata['name']
            del mdata['name']
            return cls(metadata=mdata)
        else:
            raise ValueError('Unable to retrieve information from ' + name)

    # @classmethod
    # def from_nc():
    #     """not implemented yet, load metadata from netCDF"""
    #     pass
    #
    # @classmethod
    # def from_dict():
    #     """not implemented yet, load metadata from dict of items/list types
    #     """
    #     pass



import numpy as np
import warnings

from nose.tools import raises
import pandas as pds

import pysat


class TestBasics():
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument('pysat', 'testing', tag='10',
                                         clean_level='clean')

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst

    def add(self, function, kind='add', at_pos='end', *args, **kwargs):
        '''Adds a function to the object's custom queue'''
        self.testInst.custom.add(function, kind, at_pos, *args, **kwargs)

    @raises(ValueError)
    def test_single_modifying_custom_function(self):
        """Test if custom function works correctly. Modify function that
        returns pandas object. Modify function returns an object which will
        produce an Error.
        """
        def custom1(inst):
            inst.data['doubleMLT'] = 2.0 * inst.data.mlt
            return 5.0 * inst.data['mlt']

        self.testInst.custom.add(custom1, 'modify')
        self.testInst.load(2009, 1)

    def test_single_adding_custom_function(self):
        """Test if custom function works correctly. Add function that returns
        pandas object.
        """
        def custom1(inst):
            d = 2.0 * inst['mlt']
            d.name = 'doubleMLT'
            return d

        self.add(custom1, 'add')
        self.testInst.load(2009, 1)
        assert (self.testInst['doubleMLT'].values == 2.0 *
                self.testInst['mlt'].values).all()

    def test_single_adding_custom_function_wrong_times(self):
        """Only the data at the correct time should be accepted, otherwise it
        returns nan
        """
        def custom1(inst):
            new_index = inst.index+pds.DateOffset(milliseconds=500)
            d = pds.Series(2.0 * inst['mlt'], index=new_index)
            d.name = 'doubleMLT'
            print(new_index)
            return d

        self.add(custom1, 'add')
        self.testInst.load(2009, 1)
        ans = (self.testInst['doubleMLT'].isnull()).all()
        if self.testInst.pandas_format:
            assert ans
        else:
            print("Warning! Xarray doesn't enforce the same times on all " +
                  "parameters in dataset.")

    def test_single_adding_custom_function_that_modifies_passed_data(self):
        """Test if custom function works correctly. Add function that returns
        pandas object but modifies passed satellite object.
        Changes to passed object should not propagate back.
        """
        def custom1(inst):
            inst.data['doubleMLT'] = 2.0 * inst.data.mlt
            inst['mlt'] = 0.
            return inst.data.doubleMLT

        self.add(custom1, 'add')
        self.testInst.load(2009, 1)
        assert (self.testInst.data['doubleMLT'] == 2.0 *
                self.testInst['mlt']).all()

    def test_add_function_tuple_return_style(self):
        """Test if custom function works correctly. Add function that returns
        name and numpy array.
        """
        def custom1(inst):
            return ('doubleMLT', 2.0 * inst.data.mlt.values)
        self.testInst.custom.add(custom1, 'add')
        self.testInst.load(2009, 1)
        print(self.testInst['doubleMLT'])
        print(2.0 * self.testInst['mlt'])
        assert (self.testInst['doubleMLT'] == 2.0 * self.testInst['mlt']).all()

    def test_add_multiple_custom_functions_tuple_return_style(self):
        """Test if multiple custom functions that add data work correctly. Add
        function that returns name and numpy array.
        """
        def custom1(inst):
            return (['doubleMLT', 'tripleMLT'], [2.0 * inst.data.mlt.values,
                                                 3.0 * inst.data.mlt.values])
        self.testInst.custom.add(custom1, 'add')
        self.testInst.load(2009, 1)
        assert (self.testInst.data['doubleMLT'] == 2.0 *
                self.testInst['mlt']).all()
        assert (self.testInst.data['tripleMLT'] == 3.0 *
                self.testInst['mlt']).all()

    @raises(ValueError)
    def test_add_function_tuple_return_style_too_few_elements(self):
        """Test if custom function works correctly. Add function that returns
        name and numpy array.
        """
        def custom1(inst):
            return ('doubleMLT', 2.0 * inst.data.mlt.values[0:-5])
        self.testInst.custom.add(custom1, 'add')
        self.testInst.load(2009, 1)
        if not self.testInst.pandas_format:
            print("Warning! Xarray doesn't enforce the same number of " +
                  "elements on all parameters in dataset.")
            raise ValueError

    @raises(ValueError)
    def test_add_function_tuple_return_style_too_many_elements(self):
        """Test if custom function works correctly. Add function that returns
        name and numpy array.
        """
        def custom1(inst):
            return ('doubleMLT', np.arange(2.0 * len(inst.data.mlt)))
        self.testInst.custom.add(custom1, 'add')
        self.testInst.load(2009, 1)
        if not self.testInst.pandas_format:
            print("Warning! Xarray doesn't enforce the same number of " +
                  "elements on all parameters in dataset.")
            raise ValueError

    def test_add_dataframe(self):
        def custom1(inst):
            out = pysat.DataFrame({'doubleMLT': inst.data.mlt * 2,
                                   'tripleMLT': inst.data.mlt * 3},
                                  index=inst.index)
            return out
        self.add(custom1, 'add')
        self.testInst.load(2009, 1)
        assert (self.testInst.data['doubleMLT'] == 2.0 *
                self.testInst['mlt']).all()
        assert (self.testInst.data['tripleMLT'] == 3.0 *
                self.testInst['mlt']).all()

    def test_add_dataframe_w_meta(self):
        def custom1(inst):
            out = pysat.DataFrame({'doubleMLT': inst.data.mlt * 2,
                                   'tripleMLT': inst.data.mlt * 3},
                                  index=inst.index)
            return {'data': out,
                    'long_name': ['doubleMLTlong', 'tripleMLTlong'],
                    'units': ['hours1', 'hours2']}
        self.add(custom1, 'add')
        self.testInst.load(2009, 1)
        assert self.testInst.meta['doubleMLT'].units == 'hours1'
        assert self.testInst.meta['doubleMLT'].long_name == 'doubleMLTlong'
        assert self.testInst.meta['tripleMLT'].units == 'hours2'
        assert self.testInst.meta['tripleMLT'].long_name == 'tripleMLTlong'
        assert (self.testInst['doubleMLT'] == 2.0 * self.testInst['mlt']).all()
        assert (self.testInst['tripleMLT'] == 3.0 * self.testInst['mlt']).all()

    def test_add_series_w_meta(self):
        def custom1(inst):
            out = pysat.Series(inst.data.mlt * 2,
                               index=inst.index)
            out.name = 'doubleMLT'
            return {'data': out, 'long_name': 'doubleMLTlong',
                    'units': 'hours1'}
        self.add(custom1, 'add')
        self.testInst.load(2009, 1)
        assert self.testInst.meta['doubleMLT'].units == 'hours1'
        assert self.testInst.meta['doubleMLT'].long_name == 'doubleMLTlong'
        assert (self.testInst['doubleMLT'] == 2.0 * self.testInst['mlt']).all()

    def test_add_series_w_meta_missing_long_name(self):
        def custom1(inst):
            out = pysat.Series(2.0 * inst.data.mlt.values,
                               index=inst.index)
            out.name = 'doubleMLT'
            return {'data': out,
                    'units': 'hours1'}
        self.add(custom1, 'add')
        self.testInst.load(2009, 1)
        assert self.testInst.meta['doubleMLT'].units == 'hours1'
        assert self.testInst.meta['doubleMLT'].long_name == 'doubleMLT'
        assert (self.testInst['doubleMLT'] == 2.0 * self.testInst['mlt']).all()

    def test_add_series_w_meta_name_in_dict(self):
        def custom1(inst):
            out = pysat.Series(2.0 * inst.data.mlt.values,
                               index=inst.index)
            return {'data': out, 'long_name': 'doubleMLTlong',
                    'units': 'hours1', 'name': 'doubleMLT'}
        self.add(custom1, 'add')
        self.testInst.load(2009, 1)
        assert self.testInst.meta['doubleMLT'].units == 'hours1'
        assert self.testInst.meta['doubleMLT'].long_name == 'doubleMLTlong'
        assert (self.testInst['doubleMLT'] == 2.0 * self.testInst['mlt']).all()

    @raises(ValueError)
    def test_add_series_w_meta_no_name(self):
        def custom1(inst):
            out = pysat.Series({'doubleMLT': inst.data.mlt * 2},
                               index=inst.index)
            # out.name = 'doubleMLT'
            return {'data': out, 'long_name': 'doubleMLTlong',
                    'units': 'hours1'}
        self.add(custom1, 'add')
        self.testInst.load(2009, 1)

    def test_add_numpy_array_w_meta_name_in_dict(self):
        def custom1(inst):
            out = 2.*inst['mlt'].values
            return {'data': out, 'long_name': 'doubleMLTlong',
                    'units': 'hours1', 'name': 'doubleMLT'}
        self.add(custom1, 'add')
        self.testInst.load(2009, 1)
        assert self.testInst.meta['doubleMLT'].units == 'hours1'
        assert self.testInst.meta['doubleMLT'].long_name == 'doubleMLTlong'
        assert (self.testInst['doubleMLT'] == 2.0 * self.testInst['mlt']).all()

    @raises(ValueError)
    def test_add_numpy_array_w_meta_no_name_in_dict(self):
        def custom1(inst):
            out = (inst.data.mlt * 2).values
            return {'data': out, 'long_name': 'doubleMLTlong',
                    'units': 'hours1'}
        self.add(custom1, 'add')
        self.testInst.load(2009, 1)

    def test_add_list_w_meta_name_in_dict(self):
        def custom1(inst):
            out = (inst.data.mlt * 2).values.tolist()
            return {'data': out, 'long_name': 'doubleMLTlong',
                    'units': 'hours1', 'name': 'doubleMLT'}
        self.add(custom1, 'add')
        self.testInst.load(2009, 1)
        assert self.testInst.meta['doubleMLT'].units == 'hours1'
        assert self.testInst.meta['doubleMLT'].long_name == 'doubleMLTlong'
        assert (self.testInst['doubleMLT'] == 2.0 * self.testInst['mlt']).all()

    @raises(ValueError)
    def test_add_list_w_meta_no_name_in_dict(self):
        def custom1(inst):
            out = (inst.data.mlt * 2).values.tolist()
            return {'data': out, 'long_name': 'doubleMLTlong',
                    'units': 'hours1'}
        self.testInst.custom.add(custom1, 'add')
        self.testInst.load(2009, 1)

    def test_clear_functions(self):
        def custom1(inst):
            out = (inst.data.mlt * 2).values
            return {'data': out, 'long_name': 'doubleMLTlong',
                    'units': 'hours1', 'name': 'doubleMLT'}
        self.testInst.custom.add(custom1, 'add')
        self.testInst.custom.clear()
        assert self.testInst.custom._functions == []
        assert self.testInst.custom._kind == []

    def test_pass_functions(self):
        def custom1(inst):
            out = (inst.data.mlt * 2).values
            return
        self.testInst.custom.add(custom1, 'pass')
        self.testInst.load(2009, 1)

        assert True

    @raises(ValueError)
    def test_pass_functions_no_return_allowed(self):
        def custom1(inst):
            out = (inst.data.mlt * 2).values
            return {'data': out, 'long_name': 'doubleMLTlong',
                    'units': 'hours1', 'name': 'doubleMLT'}
        self.testInst.custom.add(custom1, 'pass')
        self.testInst.load(2009, 1)

    @raises(AttributeError)
    def test_add_multiple_functions_one_not_at_end(self):
        def custom1(inst):
            out = (inst.data.mlt * 2).values
            return {'data': out, 'long_name': 'doubleMLTlong',
                    'units': 'hours1', 'name': 'doubleMLT'}

        def custom2(inst):
            out = (inst.data.mlt * 3).values
            return {'data': out, 'long_name': 'tripleMLTlong',
                    'units': 'hours1', 'name': 'tripleMLT'}

        def custom3(inst):
            out = (inst.data.tripleMLT * 2).values
            return {'data': out, 'long_name': 'quadMLTlong',
                    'units': 'hours1', 'name': 'quadMLT'}
        self.testInst.custom.add(custom1, 'add')
        self.testInst.custom.add(custom2, 'add')
        # if this runs correctly, an error will be thrown
        # since the data required by custom3 won't be present yet
        self.testInst.custom.add(custom3, 'add', at_pos=1)
        self.testInst.load(2009, 1)


class TestBasicsXarray(TestBasics):
    def setup(self):
        """Runs before every method to create a clean testing setup."""
        self.testInst = pysat.Instrument('pysat', 'testing_xarray', tag='10',
                                         clean_level='clean')

    def teardown(self):
        """Runs after every method to clean up previous testing."""
        del self.testInst


class ConstellationTestBasics(TestBasics):
    def setup(self):
        """Runs before every method to create a clean testing setup"""
        insts = []
        for i in range(5):
            insts.append(pysat.Instrument('pysat', 'testing', tag='10',
                                          clean_level='clean'))

        self.testConst = pysat.Constellation(insts)

    def teardown(self):
        """ Runs after every method to clean up previous testing"""
        del self.testConst

    def add(self, function, kind='add', at_pos='end', *args, **kwargs):
        """ Add a function to the object's custom queue"""
        self.testConst.data_mod(function, kind, at_pos, *args, **kwargs)

class TestDeprecation():

    def setup(self):
        """Runs before every method to create a clean testing setup"""
        warnings.simplefilter("always")

    def teardown(self):
        """Runs after every method to clean up previous testing"""

    def test_deprecation_warning_custom_add(self):
        """Test if custom.add is deprecated"""

        def func():
            """Fake function to attach"""
            print('Hi!')

        testInst = pysat.Instrument(platform='pysat', name='testing')
        with warnings.catch_warnings(record=True) as war:
            try:
                testInst.custom.add(func)
            except AttributeError:
                # Setting inst to None should produce a AttributeError after
                # warning is generated
                pass

        assert len(war) >= 1
        assert war[0].category == DeprecationWarning



#!/usr/bin/env python
# Full license can be found in License.md
# Full author list can be found in .zenodo.json file
# DOI:10.5281/zenodo.1199703
# ----------------------------------------------------------------------------

import collections
import glob
import numpy as np
import os
import re
import shutil
import string

import pandas as pds

from pysat.utils.time import create_datetime_index
from pysat.utils._core import available_instruments


def process_parsed_filenames(stored, two_digit_year_break=None):
    """Create a Files pandas Series of filenames from a formatted dict

    Parameters
    ----------
    stored : collections.orderedDict
        Dict produced by parse_fixed_width_filenames or
        parse_delimited_filenames
    two_digit_year_break : int or NoneType
        If filenames only store two digits for the year, then
        '1900' will be added for years >= two_digit_year_break
        and '2000' will be added for years < two_digit_year_break.
        If None, then four-digit years are assumed.
        (default=None)

    Returns
    -------
    pds.Series
        Series, indexed by datetime, with file strings

    Note
    ----
    If two files have the same date and time information in the filename then
    the file with the higher version/revision/cycle is used. Series returned
    only has one file per datetime. Version is required for this filtering,
    revision and cycle are optional.

    """

    search_dict = construct_searchstring_from_format(stored['format_str'])
    keys = search_dict['keys']

    if len(stored['files']) > 0:
        # Deal with the possibility of two digit years. Years above
        # or equal to break are considered to be 1900+, while
        # years below break are considered to be 2000+
        if two_digit_year_break is not None:
            idx, = np.where(np.array(stored['year'])
                            >= two_digit_year_break)
            stored['year'][idx] = stored['year'][idx] + 1900
            idx, = np.where(np.array(stored['year']) < two_digit_year_break)
            stored['year'][idx] = stored['year'][idx] + 2000

        # Need to sort the information for things to work
        rec_arr = [stored[key] for key in keys]
        rec_arr.append(stored['files'])

        # Sort all arrays by creating a sortable records array
        # withs keys corresponding to the files
        val_keys = keys + ['files']
        rec_arr = np.rec.fromarrays(rec_arr, names=val_keys)
        rec_arr.sort(order=val_keys, axis=0)

        # pull out sorted info
        for key in keys:
            stored[key] = rec_arr[key]
        files = rec_arr['files']

        # add hour and minute information to 'second'
        if stored['second'] is None:
            stored['second'] = np.zeros(len(files))
        if stored['hour'] is not None:
            stored['second'] += 3600 * stored['hour']
        if stored['minute'] is not None:
            stored['second'] += 60 * stored['minute']
        # version shouldn't be set to zero
        # version is required to remove duplicate datetimes
        if stored['revision'] is None:
            stored['revision'] = np.zeros(len(files))
        if stored['cycle'] is None:
            stored['cycle'] = np.zeros(len(files))

        index = create_datetime_index(year=stored['year'],
                                      month=stored['month'],
                                      day=stored['day'],
                                      uts=stored['second'])

        # if version, revision, and cycle are supplied
        # use these parameters to weed out files that have been replaced
        # with updated versions
        # first, check for duplicate index times
        dups = index[index.duplicated()].unique()
        if (len(dups) > 0) and (stored['version'] is not None):
            # we have duplicates
            # keep the highest version/revision combo
            version = pds.Series(stored['version'], index=index)
            revision = pds.Series(stored['revision'], index=index)
            cycle = pds.Series(stored['cycle'], index=index)
            revive = version * 100000. + revision + cycle * 1e-5
            frame = pds.DataFrame({'files': files, 'revive': revive,
                                   'time': index}, index=index)
            frame = frame.sort_values(by=['time', 'revive'],
                                      ascending=[True, False])
            frame = frame.drop_duplicates(subset='time', keep='first')

            return frame['files']
        else:
            return pds.Series(files, index=index)
    else:
        return pds.Series(None, dtype='object')


def parse_fixed_width_filenames(files, format_str):
    """Parses list of files, extracting data identified by format_str

    Parameters
    ----------
    files : list
        List of files
    format_str : str
        Provides the naming pattern of the instrument files and the
        locations of date information so an ordered list may be produced.
        Supports string formatting codes 'year', 'month', 'day', 'hour',
        'minute', 'second', 'version', 'revision', and 'cycle'. For example,
        `instrument_{year:4d}{month:02d}{day:02d}_v{version:02d}.cdf`

    Returns
    -------
    stored : collections.OrderedDict
        Information parsed from filenames that inclues: 'year', 'month', 'day',
        'hour', 'minute', 'second', 'version', 'revision', and 'cycle'.  Also
        includes 'files', an input list of files, and 'format_str', a formatted
        string from input

    """

    # create storage for data to be parsed from filenames
    ordered_keys = ['year', 'month', 'day', 'hour', 'minute', 'second',
                    'version', 'revision', 'cycle']
    stored = collections.OrderedDict({kk: list() for kk in ordered_keys})

    if len(files) == 0:
        stored['files'] = []
        # include format string as convenience for later functions
        stored['format_str'] = format_str
        return stored

    # parse format string to get information needed to parse filenames
    search_dict = construct_searchstring_from_format(format_str)
    snips = search_dict['string_blocks']
    lengths = search_dict['lengths']
    keys = search_dict['keys']

    # determine the locations the date/version information in a filename is
    # stored use these indices to slice out date from filenames
    idx = 0
    begin_key = []
    end_key = []
    for i, snip in enumerate(snips):
        idx += len(snip)
        if i < (len(lengths)):
            begin_key.append(idx)
            idx += lengths[i]
            end_key.append(idx)
    max_len = idx
    # setting up negative indexing to pick out filenames
    key_str_idx = [np.array(begin_key, dtype=int) - max_len,
                   np.array(end_key, dtype=int) - max_len]
    # need to parse out dates for datetime index
    for i, temp in enumerate(files):
        for j, key in enumerate(keys):
            val = temp[key_str_idx[0][j]:key_str_idx[1][j]]
            stored[key].append(val)
    # convert to numpy arrays
    for key in stored.keys():
        stored[key] = np.array(stored[key]).astype(int)
        if len(stored[key]) == 0:
            stored[key] = None
    # include files in output
    stored['files'] = files
    # include format string as convenience for later functions
    stored['format_str'] = format_str

    return stored


def parse_delimited_filenames(files, format_str, delimiter):
    """Parses list of files, extracting data identified by format_str

    Parameters
    ----------
    files : list
        List of files
    format_str : str
        Provides the naming pattern of the instrument files and the
        locations of date information so an ordered list may be produced.
        Supports string formatting codes 'year', 'month', 'day', 'hour',
        'minute', 'second', 'version', 'revision', and 'cycle'. For example,
        `instrument_{year:4d}{month:02d}{day:02d}_v{version:02d}.cdf`
    delimiter : string
        Delimiter string upon which files will be split (e.g., '.')

    Returns
    -------
    stored : collections.OrderedDict
        Information parsed from filenames that inclues: 'year', 'month', 'day',
        'hour', 'minute', 'second', 'version', 'revision', and 'cycle'.  Also
        includes 'files', an input list of files, and 'format_str', a formatted
        string from input

    """

    # create storage for data to be parsed from filenames
    ordered_keys = ['year', 'month', 'day', 'hour', 'minute', 'second',
                    'version', 'revision', 'cycle']
    stored = collections.OrderedDict({kk: list() for kk in ordered_keys})

    # exit early if there are no files
    if len(files) == 0:
        stored['files'] = []
        # include format string as convenience for later functions
        stored['format_str'] = format_str
        return stored

    # parse format string to get information needed to parse filenames
    search_dict = construct_searchstring_from_format(format_str, wildcard=True)
    snips = search_dict['string_blocks']
    keys = search_dict['keys']

    # going to parse string on delimiter
    # it is possible that other regions have the delimiter but aren't
    # going to be parsed out
    # so apply delimiter breakdown to the string blocks as a guide
    pblock = []
    parsed_block = [snip.split(delimiter) for snip in snips]
    for block in parsed_block:
        if block != ['', '']:
            if block[0] == '':
                block = block[1:]
            if block[-1] == '':
                block = block[:-1]
            pblock.extend(block)
        pblock.append('')
    parsed_block = pblock[:-1]

    # need to parse out dates for datetime index
    for temp in files:
        split_name = temp.split(delimiter)
        idx = 0
        for sname, bname in zip(split_name, parsed_block):
            if bname == '':
                # areas with data to be parsed are indicated with a
                # '' in parsed_block
                stored[keys[idx]].append(sname)
                idx += 1

    # convert to numpy arrays
    for key in stored.keys():
        try:
            # Assume key value is numeric integer
            stored[key] = np.array(stored[key]).astype(int)
        except ValueError:
            # Store key value as string
            stored[key] = np.array(stored[key])
        if len(stored[key]) == 0:
            stored[key] = None

    # include files in output
    stored['files'] = files

    # include format string as convenience for later functions
    stored['format_str'] = format_str

    return stored


def construct_searchstring_from_format(format_str, wildcard=False):
    """Parses format file string and returns string formatted for searching.

    Parameters
    ----------
    format_str : str
        Provides the naming pattern of the instrument files and the
        locations of date information so an ordered list may be produced.
        Supports 'year', 'month', 'day', 'hour', 'minute', 'second', 'version',
        'revision', and 'cycle'. For example,
        `instrument_{year:04d}{month:02d}{day:02d}_v{version:02d}.cdf`
    wildcard : bool
        If True, replaces the ? sequence with a * . This option may be well
        suited when dealing with delimited filenames.

    Returns
    -------
    out_dict : dict
        An output dict with the following keys:
        - 'search_string' (format_str with data to be parsed replaced with ?)
        - 'keys' (keys for data to be parsed)
        - 'lengths' (string length for data to be parsed)
        - 'string_blocks' (the filenames are broken into fixed width segments).

    Note
    ----
    The '?' may be used to indicate a set number of spaces for a variable
    part of the name that need not be extracted.
    `cnofs_cindi_ivm_500ms_{year:4d}{month:02d}{day:02d}_v??.cdf`

    A standards compliant filename can be constructed by starting with
    string_blocks, adding keys in order, and replacing the '' locations
    with data of length length.

    """

    out_dict = {'search_string': '',
                'keys': [],
                'lengths': [],
                'string_blocks': []}
    if format_str is None:
        raise ValueError("Must supply a filename template (format_str).")

    # Parse format string to figure out how to construct the search string
    # to identify files in the filesystem
    form = string.Formatter()
    for snip in form.parse(format_str):
        # Collect all of the format keywords. Replace them in the string with
        # the '?' wildcard. The numnber of '?'s corresponds to the length of
        # data to be parsed. The length is obtained from format keywords so
        # that we know later on where to parse information out from.
        out_dict['search_string'] += snip[0]
        out_dict['string_blocks'].append(snip[0])

        if snip[1] is not None:
            out_dict['keys'].append(snip[1])

            # Try and determine formatting width
            temp = re.findall(r'\d+', snip[2])

            if temp:
                # There are items, try and grab width
                for i in temp:
                    # Make sure there is truly something there
                    if i != 0:
                        # Store length and add to the search string
                        out_dict['lengths'].append(int(i))
                        if not wildcard:
                            out_dict['search_string'] += '?' * int(i)
                        else:
                            out_dict['search_string'] += '*'
                        break
            else:
                raise ValueError("Couldn't determine formatting width")

    return out_dict


def search_local_system_formatted_filename(data_path, search_str):
    """
    Parses format file string and returns string formatted for searching.

    Parameters
    ----------
    data_path : string
        Top level directory to search files for. This directory
        is provided by pysat to the instrument_module.list_files
        functions as data_path.
    search_str : string
        String used to search for local files. For example,
        `cnofs_cindi_ivm_500ms_????????_v??.cdf` or `inst-name-*-v??.cdf`

    Returns
    -------
    files : list
        list of files matching the specified file format

    Note
    ----
    The use of ?s (1 ? per character) rather than the full wildcard * provides
    a more specific filename search string that limits the false positive rate.

    """

    # Perform local file search
    abs_search_str = os.path.join(data_path, search_str)
    files = glob.glob(abs_search_str)

    # Remove the specified data_path portion
    files = [sfile.split(data_path)[-1] for sfile in files]

    # Return the desired filename information
    return files


def update_data_directory_structure(new_template, test_run=True,
                                    full_breakdown=False,
                                    remove_empty_dirs=False):
    """Update pysat data directory structure to match supplied template

    Translates all of pysat's managed science files to a new
    directory structure. By default, pysat uses the template string stored in
    pysat.params['directory_format'] to organize files. This method makes
    it possible to transition an existing pysat installation so it works
    with the supplied new template.

    Parameters
    ----------
    new_template : str
        New directory template string. The default value for pysat is
         `os.path.join(('{platform}', '{name}', '{tag}', '{inst_id}'))`
    test_run : bool
        If True, a printout of all proposed changes will be made, but the
        directory changes will not be enacted. (default=True)
    full_breakdown : bool
        If True, a full path for every file is printed to terminal.
        (default=False)
    remove_empty_dirs : bool
        If True, all directories that had pysat.Instrument data moved
        to another location and are now empty are deleted. Traverses
        the directory chain up to the top-level directories in
        `pysat.params['data_dirs']`. (default=False)

    Note
    ----
    After updating the data directory structures users should nominally
    assign `new_template` as the directory format via
    ::

        pysat.params['directory_format'] = new_template

    """

    # Import required here to avoid circular import
    from pysat import Instrument

    # Get a list of supported instruments
    # Best solved with an upcoming method in pull #633
    insts = available_instruments()

    if test_run:
        print('Performing a test run. No files will be moved.\n')

    if len(insts.keys()) == 0:
        ostr = ''.join(('No registered instruments detected. Please register ',
                        'supporting instrument modules using the ',
                        '`pysat.utils.registry` module.'))
        print(ostr)

    for platform in insts.keys():
        for name in insts[platform].keys():
            for inst_id in insts[platform][name]['inst_ids_tags'].keys():
                tkeys = insts[platform][name]['inst_ids_tags'][inst_id].keys()
                for tag in tkeys:

                    # Get a list of current Instrument files. Some Instruments
                    # construct filenames so we start by taking only the
                    # unique ones.
                    inst = Instrument(platform=platform, name=name,
                                      inst_id=inst_id, tag=tag,
                                      update_files=True)
                    flist = inst.files.files.values
                    flist = np.unique(flist)

                    # Existing full directory path, including template.
                    curr_path = inst.files.data_path

                    # Figure out which top-level directory these files are in by
                    # removing existing directory template information.
                    # pysat supports multiple paths so we need to get
                    # which of those this particular instrument uses.
                    currdir = inst.files.data_path.split(
                        inst.files.sub_dir_path)[0]

                    # Create instrument with new information
                    new_inst = Instrument(platform=platform, name=name,
                                          inst_id=inst_id, tag=tag,
                                          update_files=False,
                                          directory_format=new_template)

                    # Get new formatted directory template
                    subdir = new_inst.files.sub_dir_path

                    # Make new path using correct top_level data directory but
                    # with new template. new_inst won't find files and thus
                    # defaults to the first of all pysat data_dirs though
                    # we know better.
                    new_path = os.path.join(currdir, subdir, '')
                    print(' '.join(('Working on Instrument:', platform, name,
                                    tag, inst_id)))
                    if curr_path == new_path:
                        print('No change in directory needed.\n')
                        break

                    print('Current path :  ' + curr_path)
                    print('Proposed path:  ' + new_path)

                    # Construct full paths in lists for old and new filenames
                    old_files = [os.path.join(inst.files.data_path, ifile)
                                 for ifile in flist]

                    # Determine which of these files exist.
                    old_exists = [os.path.isfile(ofile) for ofile in old_files]
                    idx, = np.where(old_exists)

                    if len(idx) == 0:
                        # If none of the files actually exists, likely that
                        # instruments.methods.general.list_files is appending
                        # a date to the end of the filename.
                        exists = [os.path.isfile(ofile[:-11]) for ofile in
                                  old_files]
                        if np.all(exists):
                            flist = [ifile[:-11] for ifile in flist]
                            flist = np.unique(flist)
                            old_files = [os.path.join(inst.files.data_path,
                                                      ifile)
                                         for ifile in flist]
                        else:
                            # Files don't exist as written and taking of
                            # a trailing date didn't fix everything.
                            raise ValueError()
                    elif len(idx) < len(old_files):
                        ostr = ' '.join(('{:d} out of {:d} expected files',
                                         'were not found. It is likely',
                                         'that', platform, name,
                                         'is using a combination of internally',
                                         'modified file names and regular ',
                                         'names and must be moved manually.\n'))
                        ostr = ostr.format(len(idx), len(old_files))
                        print(ostr)
                        break

                    # User feedback
                    if len(old_files) == 0:
                        print('No files found.\n')
                    else:
                        print('{:d} files located.\n'.format(
                            len(old_files)))

                    # Based on the files that do exist, construct new
                    # path names with the updated directory template.
                    new_files = [os.path.join(currdir, subdir, ifile)
                                 for ifile in flist]

                    # Some instruments may have additional directories below
                    # those expected by pysat. Get a list of all those unique
                    # directories and create as needed.
                    new_dirs = [os.path.split(ifile)[0] for ifile in new_files]
                    new_dirs = np.unique(new_dirs)
                    if not test_run:
                        for ndir in new_dirs:
                            # Create needed directories if actually moving files
                            check_and_make_path(ndir)

                    # Ready to iterate through list of files and move them
                    for ofile, nfile in zip(old_files, new_files):
                        if full_breakdown:
                            # Print the proposed changes so user may verify
                            ostr = ''.join(('Will move: ', ofile, '\n',
                                            '       to: ', nfile))
                            print(ostr)

                        if not test_run:
                            # Move files if not in test mode
                            if not os.path.isfile(nfile):
                                # Move the file now
                                shutil.move(ofile, nfile)
                            else:
                                # New file
                                ostr = ''.join(nfile, ' already exists.')
                                print(ostr)

                    if full_breakdown and (len(old_files) > 0):
                        # Sometimes include a newline to maintain consistent
                        # line spacing.
                        print('')

                    if len(old_files) > 0:
                        # No missing files and there are actually
                        # files on the disk to deal with.

                        # Get new list of files from Instrument
                        new_inst = Instrument(platform=platform, name=name,
                                              inst_id=inst_id, tag=tag,
                                              update_files=True,
                                              directory_format=new_template)

                        # Check on number of files before and after
                        nnew = len(new_inst.files.files)
                        nold = len(inst.files.files)
                        if not test_run:
                            if nnew != nold:
                                estr = ' '.join(('Number of files before and',
                                                 'after not the same.',
                                                 'Something has gone wrong for',
                                                 platform, name, tag, inst_id))
                                raise ValueError(estr)

                            print(' '.join(('All', platform, name, tag, inst_id,
                                            'files moved and accounted for.',
                                            '\n')))

                            # Number of files checks out. Time to remove old
                            # directories if there are no real files in there.
                            # First, get full directory path of previous inst
                            wpath = inst.files.data_path
                            while wpath != currdir and (len(wpath)
                                                        > len(currdir)):
                                # Only continue while we are at a level
                                # lower than the top-level pysat data directory.
                                if len(os.listdir(wpath)) == 0:
                                    # Directory is empty, remove it.
                                    print(''.join((wpath, ' is empty and ',
                                                   'could be removed.')))
                                    if remove_empty_dirs:
                                        shutil.rmtree(wpath)
                                        print(''.join(('Removing: ', wpath)))
                                else:
                                    print(''.join(('Directory is not empty: ',
                                                   wpath, '\nEnding cleanup.',
                                                   '\n')))
                                    break

                                # Take off last path and start working up
                                # the directory chain.
                                wpath = os.path.sep.join(wpath.split(
                                    os.path.sep)[:-2])
                            else:
                                print('\n')

    return


def check_and_make_path(path):
    """Checks if path exists, creates it if it doesn't.

    Parameters
    ----------
    path : string
        Directory path without any file names. Creates all
        necessary directories to complete the path.

    Raises
    ------
    ValueError
        If input path and internally constructed path are not equal.

    """

    if not os.path.exists(path):
        # Make path, checking to see that each level exists before attempting
        root_path, local_dir = os.path.split(path)

        # Check that we have a remotely valid path
        if len(root_path) == 0:
            raise ValueError('Invalid path specification.')

        # Iterate through given path until we hit a directory that exists.
        make_dir = list()
        while not os.path.exists(root_path):
            if len(local_dir) > 0:
                # Avoid case where input is path='/stuff/level/'.
                # The trailing '/' leads to a local_dir=''
                make_dir.append(local_dir)
            root_path, local_dir = os.path.split(root_path)

        if len(local_dir) > 0:
            # Avoid case where input is path='/stuff/level/'.
            # The trailing '/' leads to a local_dir=''
            make_dir.append(local_dir)

        while len(make_dir) > 0:
            local_dir = make_dir.pop()
            root_path = os.path.join(root_path, local_dir)
            os.mkdir(root_path)

        if os.path.normpath(root_path) != os.path.normpath(path):
            raise ValueError('Desired and constructed paths differ')

    return



# -*- coding: utf-8 -*-.
"""Provides default routines for integrating CEDAR Madrigal instruments into
pysat, reducing the amount of user intervention.

 """

from __future__ import absolute_import
from __future__ import print_function

import datetime as dt
import logging
import numpy as np
import os
import pandas as pds
import subprocess
import sys

import h5py
from madrigalWeb import madrigalWeb

import pysat

logger = logging.getLogger(__name__)

def cedar_rules():
    """ General acknowledgement statement for Madrigal data.

    Returns
    -------
    ackn : string
        String with general acknowledgement for all CEDAR Madrigal data

    """
    ackn = "Contact the PI when using this data, in accordance with the CEDAR"
    ackn += " 'Rules of the Road'"
    return ackn


# support load routine
def load(fnames, tag=None, sat_id=None, xarray_coords=[]):
    """Loads data from Madrigal into Pandas.

    This routine is called as needed by pysat. It is not intended
    for direct user interaction.

    Parameters
    ----------
    fnames : array-like
        iterable of filename strings, full path, to data files to be loaded.
        This input is nominally provided by pysat itself.
    tag : string ('')
        tag name used to identify particular data set to be loaded.
        This input is nominally provided by pysat itself. While
        tag defaults to None here, pysat provides '' as the default
        tag unless specified by user at Instrument instantiation.
    sat_id : string ('')
        Satellite ID used to identify particular data set to be loaded.
        This input is nominally provided by pysat itself.
    xarray_coords : list
        List of keywords to use as coordinates if xarray output is desired
        instead of a Pandas DataFrame (default=[])

    Returns
    -------
    data : pds.DataFrame or xr.DataSet
        A pandas DataFrame or xarray DataSet holding the data from the HDF5
        file
    metadata : pysat.Meta
        Metadata from the HDF5 file, as well as default values from pysat

    Examples
    --------
    ::
        inst = pysat.Instrument('jro', 'isr', 'drifts')
        inst.load(2010,18)

    """

    # Ensure 'time' wasn't included as a coordinate, since it is the default
    if 'time' in xarray_coords:
        xarray_coords.pop(xarray_coords.index('time'))

    # Open the specified file
    filed = h5py.File(fnames[0], 'r')
    # data
    file_data = filed['Data']['Table Layout']
    # metadata
    file_meta = filed['Metadata']['Data Parameters']
    # load up what is offered into pysat.Meta
    meta = pysat.Meta()
    meta.info = {'acknowledgements': "See 'meta.Experiment_Notes' for " +
                 "instrument specific acknowledgements\n" + cedar_rules(),
                 'references': "See 'meta.Experiment_Notes' for references"}
    labels = []
    for item in file_meta:
        # handle difference in string output between python 2 and 3
        name_string = item[0]
        unit_string = item[3]
        desc_string = item[1]
        if sys.version_info[0] >= 3:
            name_string = name_string.decode('UTF-8')
            unit_string = unit_string.decode('UTF-8')
            desc_string = desc_string.decode('UTF-8')
        labels.append(name_string)
        meta[name_string.lower()] = {'long_name': name_string,
                                     'units': unit_string,
                                     'desc': desc_string}

    # add additional metadata notes
    # custom attributes attached to meta are attached to
    # corresponding Instrument object when pysat receives
    # data and meta from this routine
    for key in filed['Metadata']:
        if key != 'Data Parameters':
            setattr(meta, key.replace(' ', '_'), filed['Metadata'][key][:])
    # data into frame, with labels from metadata
    data = pds.DataFrame.from_records(file_data, columns=labels)
    # lowercase variable names
    data.columns = [item.lower() for item in data.columns]
    # datetime index from times
    time_keys = np.array(['year', 'month', 'day', 'hour', 'min', 'sec'])
    if not np.all([key in data.columns for key in time_keys]):
        time_keys = [key for key in time_keys if key not in data.columns]
        raise ValueError("unable to construct time index, missing " +
                         "{:}".format(time_keys))

    uts = 3600.0 * data.loc[:, 'hour'] + 60.0 * data.loc[:, 'min'] \
        + data.loc[:, 'sec']
    time = pysat.utils.time.create_datetime_index(year=data.loc[:, 'year'],
                                                  month=data.loc[:, 'month'],
                                                  day=data.loc[:, 'day'],
                                                  uts=uts)
    # Declare index or recast as xarray
    if len(xarray_coords) > 0:
        if not np.all([xkey.lower() in data.columns
                       for xkey in xarray_coords]):
            estr = 'unknown coordinate key in {:}, '.format(xarray_coords)
            estr += 'use only {:}'.format(data.columns)
            raise ValueError(estr)

        # Append time to the data frame and add as the first coordinate
        data = data.assign(time=pds.Series(time, index=data.index))
        xarray_coords.insert(0, 'time')

        # Set the indices
        data = data.set_index(xarray_coords)

        # Recast the data as an xarray
        data = data.to_xarray()
    else:
        # Set the index to time, and put up a warning if there are duplicate
        # times.  This could mean the data should be stored as an xarray
        # DataSet
        data.index = time

        if np.any(time.duplicated()):
            logger.warning("duplicated time indices, consider specifing " +
                  "additional coordinates and storing the data as an xarray" +
                  " DataSet")

    return data, meta


def download(date_array, inst_code=None, kindat=None, data_path=None,
             user=None, password=None, url="http://cedar.openmadrigal.org",
             file_format='hdf5'):
    """Downloads data from Madrigal.

    Parameters
    ----------
    date_array : array-like
        list of datetimes to download data for. The sequence of dates need not
        be contiguous.
    inst_code : string (None)
        Madrigal instrument code(s), cast as a string.  If multiple are used,
        separate them with commas.
    kindat : string  (None)
        Experiment instrument code(s), cast as a string.  If multiple are used,
        separate them with commas.
    data_path : string (None)
        Path to directory to download data to.
    user : string (None)
        User string input used for download. Provided by user and passed via
        pysat. If an account
        is required for dowloads this routine here must error if user not
        supplied.
    password : string (None)
        Password for data download.
    url : string ('http://cedar.openmadrigal.org')
        URL for Madrigal site
    file_format : string ('hdf5')
        File format for Madrigal data.  Load routines currently only accept
        'hdf5', but any of the Madrigal options may be used here.

    Returns
    --------
    Void : (NoneType)
        Downloads data to disk.

    Notes
    -----
    The user's names should be provided in field user. Ruby Payne-Scott should
    be entered as Ruby+Payne-Scott

    The password field should be the user's email address. These parameters
    are passed to Madrigal when downloading.

    The affiliation field is set to pysat to enable tracking of pysat
    downloads.

    """

    if inst_code is None:
        raise ValueError("Must supply Madrigal instrument code")

    if kindat is None:
        raise ValueError("Must supply Madrigal experiment code")

    # currently passes things along if no user and password supplied
    # need to do this for testing
    # TODO, implement user and password values in test code
    # specific to each instrument
    if user is None:
        logger.info('No user information supplied for download.')
        user = 'pysat_testing'
    if password is None:
        logger.info('Please provide email address in password field.')
        password = 'pysat_testing@not_real_email.org'

    # Initialize the connection to Madrigal
    web_data = madrigalWeb.MadrigalData(url)

    # Get the list of desired remote files
    start = date_array.min()
    stop = date_array.max()
    if start == stop:
        stop += dt.timedelta(days=1)
    files = get_remote_filenames(inst_code=inst_code, kindat=kindat, user=user,
                                 password=password, web_data=web_data, url=url,
                                 start=start, stop=stop)

    for mad_file in files:
        local_file = os.path.join(data_path, os.path.basename(mad_file.name))

        if not os.path.isfile(local_file):
            web_data.downloadFile(mad_file.name, local_file, user, password,
                                  "pysat", format=file_format)


def get_remote_filenames(inst_code=None, kindat=None, user=None,
                         password=None, web_data=None,
                         url="http://cedar.openmadrigal.org",
                         start=dt.datetime(1900,1,1), stop=dt.datetime.now(),
                         date_array=None):
    """Retrieve the remote filenames for a specified Madrigal instrument
    (and experiment)

    Parameters
    ----------
    inst_code : string (None)
        Madrigal instrument code(s), cast as a string.  If multiple are used,
        separate them with commas.
    kindat : string (None)
        Madrigal experiment code(s), cast as a string.  If multiple are used,
        separate them with commas.  If not supplied, all will be returned.
    data_path : string (None)
        Path to directory to download data to.
    user : string (None)
        User string input used for download. Provided by user and passed via
        pysat. If an account
        is required for dowloads this routine here must error if user not
        supplied.
    password : string (None)
        Password for data download.
    web_data : MadrigalData (None)
        Open connection to Madrigal database or None (will initiate using url)
    url : string ('http://cedar.openmadrigal.org')
        URL for Madrigal site
    start : dt.datetime
        Starting time for file list (defaults to 01-01-1900)
    stop : dt.datetime
        Ending time for the file list (defaults to time of run)
    date_array : dt.datetime (None)
        Array of datetimes to download data for. The sequence of dates need not
        be contiguous and will be used instead of start and stop if supplied.

    Returns
    -------
    Void : (NoneType)
        Downloads data to disk.

    Notes
    -----
    The user's names should be provided in field user. Ruby Payne-Scott should
    be entered as Ruby+Payne-Scott

    The password field should be the user's email address. These parameters
    are passed to Madrigal when downloading.

    The affiliation field is set to pysat to enable tracking of pysat
    downloads.


    """

    if inst_code is None:
        raise ValueError("Must supply Madrigal instrument code")

    if kindat is None:
        kindat = []
    else:
        kindat = [int(kk) for kk in kindat.split(",")]

    # currently passes things along if no user and password supplied
    # need to do this for testing
    # TODO, implement user and password values in test code
    # specific to each instrument
    if user is None:
        print('No user information supplied for download.')
        user = 'pysat_testing'
    if password is None:
        print('Please provide email address in password field.')
        password = 'pysat_testing@not_real_email.org'

    # If date_array supplied, overwrite start and stop
    if date_array is not None:
        if len(date_array) == 0:
            raise ValueError('unknown date_array supplied: {:}'.format(
                date_array))
        start = date_array.min()
        stop = date_array.max()
        if start == stop:
            stop += dt.timedelta(days=1)

    # open connection to Madrigal
    if web_data is None:
        web_data = madrigalWeb.MadrigalData(url)

    # get list of experiments for instrument from 1900 till now
    exp_list = web_data.getExperiments(inst_code, start.year, start.month,
                                       start.day, start.hour, start.minute,
                                       start.second, stop.year, stop.month,
                                       stop.day, stop.hour, stop.minute,
                                       stop.second)

    # iterate over experiments to grab files for each one
    files = list()
    print("Found {:d} Madrigral experiments".format(len(exp_list)))
    for exp in exp_list:
        if good_exp(exp, date_array=date_array):
            file_list = web_data.getExperimentFiles(exp.id)

            if len(kindat) == 0:
                files.extend(file_list)
            else:
                for file_obj in file_list:
                    if file_obj.kindat in kindat:
                        files.append(file_obj)

    return files

def good_exp(exp, date_array=None):
    """ Determine if a Madrigal experiment has good data for specified dates

    Parameters
    ----------
    exp : MadrigalExperimentFile
        MadrigalExperimentFile object
    date_array : array-like
        list of datetimes to download data for. The sequence of dates need not
        be contiguous.

    Returns
    -------
    gflag : boolean
        True if good, False if bad

    """

    gflag = False

    if exp.id != -1:
        if date_array is None:
            gflag = True
        else:
            exp_start = dt.datetime(exp.startyear, exp.startmonth, exp.startday,
                                    exp.starthour, exp.startmin, exp.startsec)
            exp_end = dt.datetime(exp.endyear, exp.endmonth, exp.endday,
                                  exp.endhour, exp.endmin, exp.endsec)

            for date_val in date_array:
                if date_val >= exp_start and date_val < exp_end:
                    gflag = True
                    break

    return gflag

def list_remote_files(tag, sat_id, inst_code=None, kindat=None, user=None,
                      password=None, supported_tags=None,
                      url="http://cedar.openmadrigal.org",
                      two_digit_year_break=None, start=dt.datetime(1900,1,1),
                      stop=dt.datetime.now()):
    """List files available from Madrigal.

    Parameters
    ----------
    tag : (string or NoneType)
        Denotes type of file to load.  Accepted types are <tag strings>.
        (default=None)
    sat_id : (string or NoneType)
        Specifies the satellite ID for a constellation.  Not used.
        (default=None)
    inst_code : string (None)
        Madrigal instrument code(s), cast as a string.  If multiple are used,
        separate them with commas.
    kindat : string (None)
        Madrigal experiment code(s), cast as a string.  If multiple are used,
        separate them with commas.  If not supplied, all will be returned.
    data_path : string (None)
        Path to directory to download data to.
    user : string (None)
        User string input used for download. Provided by user and passed via
        pysat. If an account
        is required for dowloads this routine here must error if user not
        supplied.
    password : string (None)
        Password for data download.
    supported_tags : (dict or NoneType)
        keys are sat_id, each containing a dict keyed by tag
        where the values file format template strings. (default=None)
    url : string ('http://cedar.openmadrigal.org')
        URL for Madrigal site
    two_digit_year_break : int
        If filenames only store two digits for the year, then
        '1900' will be added for years >= two_digit_year_break
        and '2000' will be added for years < two_digit_year_break.
    start : (dt.datetime)
        Starting time for file list (defaults to 01-01-1900)
    stop : (dt.datetime)
        Ending time for the file list (defaults to time of run)

    Returns
    -------
    Void : (NoneType)
        Downloads data to disk.

    Notes
    -----
    The user's names should be provided in field user. Ruby Payne-Scott should
    be entered as Ruby+Payne-Scott

    The password field should be the user's email address. These parameters
    are passed to Madrigal when downloading.

    The affiliation field is set to pysat to enable tracking of pysat
    downloads.

    Examples
    --------
    This method is intended to be set in an instrument support file at the
    top level using functools.partial
    ::
        list_remote_files = functools.partial(mad_meth.list_remote_files,
                                              supported_tags=supported_tags,
                                              inst_code=madrigal_inst_code)

    """
    if inst_code is None:
        raise ValueError("Must supply Madrigal instrument code")

    # currently passes things along if no user and password supplied
    # need to do this for testing
    # TODO, implement user and password values in test code
    # specific to each instrument
    if user is None:
        logger.info('No user information supplied for download.')
        user = 'pysat_testing'
    if password is None:
        logger.info('Please provide email address in password field.')
        password = 'pysat_testing@not_real_email.org'

    # Test input
    try:
        format_str = supported_tags[sat_id][tag]
    except KeyError:
        raise ValueError('Problem parsing supported_tags')

    # Retrieve remote file list
    files = get_remote_filenames(inst_code=inst_code, kindat=kindat, user=user,
                                 password=password, url=url, start=start,
                                 stop=stop)

    # parse these filenames to grab out the ones we want
    logger.info("Parsing filenames")
    stored = pysat._files.parse_fixed_width_filenames(files, format_str)

    # process the parsed filenames and return a properly formatted Series
    logger.info("Processing filenames")
    return pysat._files.process_parsed_filenames(stored, two_digit_year_break)


def filter_data_single_date(self):
    """Filters data to a single date.

    Parameters
    ----------
    self : pysat.Instrument
        This object

    Note
    ----
    Madrigal serves multiple days within a single JRO file
    to counter this, we will filter each loaded day so that it only
    contains the relevant day of data. This is only applied if loading
    by date. It is not applied when supplying pysat with a specific
    filename to load, nor when data padding is enabled. Note that when
    data padding is enabled the final data available within the instrument
    will be downselected by pysat to only include the date specified.

    This routine is intended to be added to the Instrument
    nanokernel processing queue via
    ::

        inst = pysat.Instrument()
        inst.custom.add(filter_data_single_date, 'modify')

    This function will then be automatically applied to the
    Instrument object data on every load by the pysat nanokernel.

    Warnings
    --------
    For the best performance, this function should be added first in the queue.
    This may be ensured by setting the default function in a
    pysat instrument file to this one.

    within platform_name.py set
    ::

        default = pysat.instruments.methods.madrigal.filter_data_single_date

    at the top level

    """

    # only do this if loading by date!
    if self._load_by_date and self.pad is None:
        # identify times for the loaded date
        idx, = np.where((self.index >= self.date) &
                        (self.index < (self.date+pds.DateOffset(days=1))))
        # downselect from all data
        self.data = self[idx]



import numpy as np
import os

import pandas as pds

import pysat


def list_files(tag=None, sat_id=None, data_path=None, format_str=None,
               file_date_range=None, test_dates=None):
    """Produce a fake list of files spanning three years

    Parameters
    ----------
    tag : (str)
        pysat instrument tag (default=None)
    sat_id : (str)
        pysat satellite ID tag (default=None)
    data_path : (str)
        pysat data path (default=None)
    format_str : (str)
        file format string (default=None)
    file_date_range : (pds.date_range)
        File date range. The default mode generates a list of 3 years of daily
        files (1 year back, 2 years forward) based on the test_dates passed
        through below.  Otherwise, accepts a range of files specified by the
        user.
        (default=None)
    test_dates : (dt.datetime)
        Pass the _test_date object through from the test instrument files

    Returns
    -------
    Series of filenames indexed by file time

    """

    # Determine the appropriate date range for the fake files
    if file_date_range is None:
        start = test_dates[''][''] - pds.DateOffset(years=1)
        stop = (test_dates[''][''] + pds.DateOffset(years=2)
                - pds.DateOffset(days=1))
        file_date_range = pds.date_range(start, stop)

    index = file_date_range

    # Create the list of fake filenames
    names = [data_path + date.strftime('%Y-%m-%d') + '.nofile'
             for date in index]

    return pysat.Series(names, index=index)


def download(date_array, tag, sat_id, data_path=None, user=None,
             password=None):
    """Simple pass function for pysat compatibility for test instruments.

    This routine is invoked by pysat and is not intended for direct use by the
    end user.

    Parameters
    ----------
    date_array : array-like
        list of datetimes to download data for. The sequence of dates need not
        be contiguous.
    tag : string ('')
        Tag identifier used for particular dataset. This input is provided by
        pysat.
    sat_id : string  ('')
        Satellite ID string identifier used for particular dataset. This input
        is provided by pysat.
    data_path : string (None)
        Path to directory to download data to.
    user : string (None)
        User string input used for download. Provided by user and passed via
        pysat. If an account
        is required for dowloads this routine here must error if user not
        supplied.
    password : string (None)
        Password for data download.
    **kwargs : dict
        Additional keywords supplied by user when invoking the download
        routine attached to a pysat.Instrument object are passed to this
        routine via kwargs.

    Returns
    --------
    Void : (NoneType)
        Downloads data to disk.

    """

    pass


def generate_fake_data(t0, num_array, period=5820, data_range=[0.0, 24.0],
                       cyclic=True):
    """Generates fake data over a given range

    Parameters
    ----------
    t0 : float
        Start time in seconds
    num_array : array_like
        Array of time steps from t0.  This is the index of the fake data
    period : int
        The number of seconds per period.
        (default = 5820)
    data_range : float
        For cyclic functions, the range of data values cycled over one period.
        Not used for non-cyclic functions.
        (default = 24.0)
    cyclic : bool
        If True, assume that fake data is a cyclic function (ie, longitude,
        slt) that will reset to data_range[0] once it reached data_range[1].
        If False, continue to monotonically increase
    """

    if cyclic:
        uts_root = np.mod(t0, period)
        data = (np.mod(uts_root + num_array, period)
                * (np.diff(data_range)[0] / float(period))) + data_range[0]
    else:
        data = ((t0 + num_array) / period).astype(int)

    return data


def generate_times(fnames, sat_id, freq='1S'):
    """Construct list of times for simulated instruments

    Parameters
    ----------
    fnames : (list)
        List of filenames.  Currently, only the first is used.  Does not
        support multi-file days as of yet.
    sat_id : (str or NoneType)
        Instrument satellite ID (accepts '' or a number (i.e., '10'), which
        specifies the number of data points to include in the test instrument)
    freq : string
        Frequency of temporal output, compatible with pandas.date_range
        [default : '1S']

    Outputs
    -------
    uts : (array)
        Array of integers representing uts for a given day
    index : (DatetimeIndex)
        The DatetimeIndex to be used in the pysat test instrument objects
    date : (datetime)
        The requested date reconstructed from the fake file name
    """

    # TODO: Expand for multi-file days
    # grab date from filename
    parts = os.path.split(fnames[0])[-1].split('-')
    yr = int(parts[0])
    month = int(parts[1])
    day = int(parts[2][0:2])
    date = pysat.datetime(yr, month, day)

    # Create one day of data at desired frequency
    index = pds.date_range(start=date, end=date+pds.DateOffset(seconds=86399),
                           freq=freq)
    # Allow numeric string to select first set of data
    try:
        index = index[0:int(sat_id)]
    except ValueError:
        # non-integer sat_id produces ValueError
        pass

    uts = index.hour*3600 + index.minute*60 + index.second

    return uts, index, date


def define_period():
    """Define the default periods for the fake data functions

    Parameters
    ----------
    None

    Returns
    -------
    period : dict
        Dictionary of periods to use in test instruments

    """

    period = {'lt': 5820, # 97 minutes
              'lon': 6240, # 104 minutes
              'angle': 5820}

    return period


def define_range():
    """Define the default ranges for the fake data functions

    Parameters
    ----------
    None

    Returns
    -------
    range : dict
        Dictionary of periods to use in test instruments

    """

    range = {'lt': [0.0, 24.0],
             'lon': [0.0, 360.0],
             'angle': [0.0, 2.0*np.pi]}

    return range
